from collections import defaultdict
from collections.abc import Iterable, Iterator, Mapping, Sequence
from datetime import datetime
from functools import cached_property
from threading import RLock
from typing import TYPE_CHECKING, AbstractSet, Callable, Optional, Union  # noqa: UP035

import dagster._check as check
from dagster import AssetSelection
from dagster._config.snap import ConfigFieldSnap, ConfigSchemaSnapshot
from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.asset_job import IMPLICIT_ASSET_JOB_NAME
from dagster._core.definitions.automation_condition_sensor_definition import (
    DEFAULT_AUTOMATION_CONDITION_SENSOR_NAME,
)
from dagster._core.definitions.backfill_policy import BackfillPolicy
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.metadata import MetadataValue
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.run_request import InstigatorType
from dagster._core.definitions.schedule_definition import DefaultScheduleStatus
from dagster._core.definitions.selector import (
    InstigatorSelector,
    RepositorySelector,
    ScheduleSelector,
    SensorSelector,
)
from dagster._core.definitions.sensor_definition import (
    DEFAULT_SENSOR_DAEMON_INTERVAL,
    DefaultSensorStatus,
    SensorType,
)
from dagster._core.definitions.utils import get_default_automation_condition_sensor_selection
from dagster._core.errors import DagsterError
from dagster._core.execution.plan.handle import ResolvedFromDynamicStepHandle, StepHandle
from dagster._core.instance import DagsterInstance
from dagster._core.origin import JobPythonOrigin, RepositoryPythonOrigin
from dagster._core.remote_representation.external_data import (
    DEFAULT_MODE_NAME,
    AssetCheckNodeSnap,
    AssetNodeSnap,
    EnvVarConsumer,
    JobDataSnap,
    JobRefSnap,
    NestedResource,
    PartitionSetSnap,
    PresetSnap,
    RepositorySnap,
    ResourceJobUsageEntry,
    ResourceSnap,
    ResourceValueSnap,
    ScheduleSnap,
    SensorMetadataSnap,
    SensorSnap,
    TargetSnap,
)
from dagster._core.remote_representation.handle import (
    CompoundID,
    InstigatorHandle,
    JobHandle,
    PartitionSetHandle,
    RepositoryHandle,
)
from dagster._core.remote_representation.job_index import JobIndex
from dagster._core.remote_representation.origin import (
    RemoteInstigatorOrigin,
    RemoteJobOrigin,
    RemotePartitionSetOrigin,
    RemoteRepositoryOrigin,
)
from dagster._core.remote_representation.represented import RepresentedJob
from dagster._core.snap import ExecutionPlanSnapshot
from dagster._core.snap.job_snapshot import JobSnap
from dagster._core.utils import toposort
from dagster._serdes import create_snapshot_id
from dagster._utils.cached_method import cached_method
from dagster._utils.schedules import schedule_execution_time_iterator

if TYPE_CHECKING:
    from dagster._core.definitions.remote_asset_graph import RemoteRepositoryAssetGraph
    from dagster._core.scheduler.instigation import InstigatorState
    from dagster._core.snap.execution_plan_snapshot import ExecutionStepSnap

_empty_set = frozenset()


class RemoteRepository:
    """RemoteRepository is a object that represents a loaded repository definition that
    is resident in another process or container. Host processes such as dagster-webserver use
    objects such as these to interact with user-defined artifacts.
    """

    def __init__(
        self,
        repository_snap: RepositorySnap,
        repository_handle: RepositoryHandle,
        instance: DagsterInstance,
        ref_to_data_fn: Optional[Callable[[JobRefSnap], JobDataSnap]] = None,
    ):
        self.repository_snap = check.inst_param(repository_snap, "repository_snap", RepositorySnap)

        self._instance = instance

        if repository_snap.job_datas is not None:
            self._job_map: dict[str, Union[JobDataSnap, JobRefSnap]] = {
                d.name: d for d in repository_snap.job_datas
            }
            self._deferred_snapshots: bool = False
            self._ref_to_data_fn = None
        elif repository_snap.job_refs is not None:
            self._job_map = {r.name: r for r in repository_snap.job_refs}
            self._deferred_snapshots = True
            if ref_to_data_fn is None:
                check.failed(
                    "ref_to_data_fn is required when RepositorySnap is loaded with deferred"
                    " snapshots"
                )

            self._ref_to_data_fn = ref_to_data_fn
        else:
            check.failed("invalid state - expected job data or refs")

        self._handle = check.inst_param(repository_handle, "repository_handle", RepositoryHandle)

        self._asset_jobs: dict[str, list[AssetNodeSnap]] = {}
        for asset_node in repository_snap.asset_nodes:
            for job_name in asset_node.job_names:
                self._asset_jobs.setdefault(job_name, []).append(asset_node)

        self._asset_check_jobs: dict[str, list[AssetCheckNodeSnap]] = {}
        for asset_check_node_snap in repository_snap.asset_check_nodes or []:
            for job_name in asset_check_node_snap.job_names:
                self._asset_check_jobs.setdefault(job_name, []).append(asset_check_node_snap)

        # memoize job instances to share instances
        self._memo_lock: RLock = RLock()
        self._cached_jobs: dict[str, RemoteJob] = {}

    @property
    def name(self) -> str:
        return self.repository_snap.name

    @property
    @cached_method
    def _schedules(self) -> dict[str, "RemoteSchedule"]:
        return {
            schedule_snap.name: RemoteSchedule(schedule_snap, self._handle)
            for schedule_snap in self.repository_snap.schedules
        }

    def has_schedule(self, schedule_name: str) -> bool:
        return schedule_name in self._schedules

    def get_schedule(self, schedule_name: str) -> "RemoteSchedule":
        return self._schedules[schedule_name]

    def get_schedules(self) -> Sequence["RemoteSchedule"]:
        return list(self._schedules.values())

    @property
    @cached_method
    def _resources(self) -> dict[str, "RemoteResource"]:
        return {
            resource_snap.name: RemoteResource(resource_snap, self._handle)
            for resource_snap in (self.repository_snap.resources or [])
        }

    def has_resource(self, resource_name: str) -> bool:
        return resource_name in self._resources

    def get_resource(self, resource_name: str) -> "RemoteResource":
        return self._resources[resource_name]

    def get_resources(self) -> Iterable["RemoteResource"]:
        return self._resources.values()

    @property
    def _utilized_env_vars(self) -> Mapping[str, Sequence[EnvVarConsumer]]:
        return self.repository_snap.utilized_env_vars or {}

    def get_utilized_env_vars(self) -> Mapping[str, Sequence[EnvVarConsumer]]:
        return self._utilized_env_vars

    @property
    @cached_method
    def _sensors(self) -> dict[str, "RemoteSensor"]:
        sensor_datas = {
            sensor_snap.name: RemoteSensor(sensor_snap, self._handle)
            for sensor_snap in self.repository_snap.sensors
        }

        if not self._instance.auto_materialize_use_sensors:
            return sensor_datas

        # if necessary, create a default automation condition sensor
        # NOTE: if a user's code location is at a version >= 1.9, then this step should
        # never be necessary, as this will be added in Definitions construction process
        default_sensor_selection = get_default_automation_condition_sensor_selection(
            sensors=[data for data in sensor_datas.values()],
            asset_graph=self.asset_graph,
        )
        if default_sensor_selection is not None:
            default_sensor_data = SensorSnap(
                name=DEFAULT_AUTOMATION_CONDITION_SENSOR_NAME,
                job_name=None,
                op_selection=None,
                asset_selection=default_sensor_selection,
                mode=None,
                min_interval=30,
                description=None,
                target_dict={},
                metadata=None,
                default_status=None,
                sensor_type=SensorType.AUTO_MATERIALIZE,
                run_tags=None,
            )
            sensor_datas[default_sensor_data.name] = RemoteSensor(default_sensor_data, self._handle)

        return sensor_datas

    def has_sensor(self, sensor_name: str) -> bool:
        return sensor_name in self._sensors

    def get_sensor(self, sensor_name: str) -> "RemoteSensor":
        return self._sensors[sensor_name]

    def get_sensors(self) -> Sequence["RemoteSensor"]:
        return list(self._sensors.values())

    @property
    @cached_method
    def _partition_sets(self) -> dict[str, "RemotePartitionSet"]:
        return {
            partition_set_snap.name: RemotePartitionSet(partition_set_snap, self._handle)
            for partition_set_snap in self.repository_snap.partition_sets
        }

    def has_partition_set(self, partition_set_name: str) -> bool:
        return partition_set_name in self._partition_sets

    def get_partition_set(self, partition_set_name: str) -> "RemotePartitionSet":
        return self._partition_sets[partition_set_name]

    def get_partition_sets(self) -> Sequence["RemotePartitionSet"]:
        return list(self._partition_sets.values())

    def has_job(self, job_name: str) -> bool:
        return job_name in self._job_map

    def get_full_job(self, job_name: str) -> "RemoteJob":
        check.str_param(job_name, "job_name")
        check.invariant(self.has_job(job_name), f'No remote job named "{job_name}" found')
        with self._memo_lock:
            if job_name not in self._cached_jobs:
                job_item = self._job_map[job_name]
                if self._deferred_snapshots:
                    if not isinstance(job_item, JobRefSnap):
                        check.failed("unexpected job item")
                    job_ref = job_item
                    job_data_snap: Optional[JobDataSnap] = None
                else:
                    if not isinstance(job_item, JobDataSnap):
                        check.failed("unexpected job item")
                    job_data_snap = job_item
                    job_ref = None

                self._cached_jobs[job_name] = RemoteJob(
                    job_data_snap=job_data_snap,
                    repository_handle=self.handle,
                    job_ref_snap=job_ref,
                    ref_to_data_fn=self._ref_to_data_fn,
                )

            return self._cached_jobs[job_name]

    def get_all_jobs(self) -> Sequence["RemoteJob"]:
        return [self.get_full_job(pn) for pn in self._job_map]

    @property
    def handle(self) -> RepositoryHandle:
        return self._handle

    @property
    def selector(self) -> RepositorySelector:
        return RepositorySelector(
            location_name=self._handle.location_name,
            repository_name=self._handle.repository_name,
        )

    @property
    def selector_id(self) -> str:
        return create_snapshot_id(self.selector)

    def get_compound_id(self) -> CompoundID:
        return CompoundID(
            remote_origin_id=self.get_remote_origin_id(),
            selector_id=self.selector_id,
        )

    def get_remote_origin(self) -> RemoteRepositoryOrigin:
        return self.handle.get_remote_origin()

    def get_python_origin(self) -> RepositoryPythonOrigin:
        return self.handle.get_python_origin()

    def get_remote_origin_id(self) -> str:
        """A means of identifying the repository this RemoteRepository represents based on
        where it came from.
        """
        return self.get_remote_origin().get_id()

    def get_asset_node_snaps(self, job_name: Optional[str] = None) -> Sequence[AssetNodeSnap]:
        return (
            self.repository_snap.asset_nodes
            if job_name is None
            else self._asset_jobs.get(job_name, [])
        )

    @cached_property
    def _asset_snaps_by_key(self) -> Mapping[AssetKey, AssetNodeSnap]:
        mapping = {}
        for asset_snap in self.repository_snap.asset_nodes:
            mapping[asset_snap.asset_key] = asset_snap
        return mapping

    def get_asset_node_snap(self, asset_key: AssetKey) -> Optional[AssetNodeSnap]:
        return self._asset_snaps_by_key.get(asset_key)

    def get_asset_check_node_snaps(
        self, job_name: Optional[str] = None
    ) -> Sequence[AssetCheckNodeSnap]:
        if job_name:
            return self._asset_check_jobs.get(job_name, [])
        else:
            return self.repository_snap.asset_check_nodes or []

    def get_display_metadata(self) -> Mapping[str, str]:
        return self.handle.display_metadata

    @cached_property
    def asset_graph(self) -> "RemoteRepositoryAssetGraph":
        """Returns a repository scoped RemoteAssetGraph."""
        from dagster._core.definitions.remote_asset_graph import RemoteRepositoryAssetGraph

        return RemoteRepositoryAssetGraph.build(self)

    def get_partition_names_for_asset_job(
        self,
        job_name: str,
        selected_asset_keys: Optional[AbstractSet[AssetKey]],
        instance: DagsterInstance,
    ) -> Sequence[str]:
        partitions_def = self._get_partitions_def_for_job(
            job_name=job_name, selected_asset_keys=selected_asset_keys
        )
        if not partitions_def:
            return []
        return partitions_def.get_partition_keys(dynamic_partitions_store=instance)

    def get_partition_tags_for_implicit_asset_job(
        self,
        job_name: str,
        selected_asset_keys: Optional[AbstractSet[AssetKey]],
        instance: DagsterInstance,
        partition_name: str,
    ) -> Mapping[str, str]:
        return check.not_none(
            self._get_partitions_def_for_job(
                job_name=job_name, selected_asset_keys=selected_asset_keys
            )
        ).get_tags_for_partition_key(partition_name)

    def _get_partitions_def_for_job(
        self,
        job_name: str,
        selected_asset_keys: Optional[AbstractSet[AssetKey]],
    ) -> Optional[PartitionsDefinition]:
        asset_nodes = self.get_asset_node_snaps(job_name)
        unique_partitions_defs: set[PartitionsDefinition] = set()
        for asset_node in asset_nodes:
            if selected_asset_keys is not None and asset_node.asset_key not in selected_asset_keys:
                continue

            if asset_node.partitions is not None:
                unique_partitions_defs.add(asset_node.partitions.get_partitions_definition())

        if len(unique_partitions_defs) == 0:
            # Assets are all unpartitioned
            return None
        if len(unique_partitions_defs) == 1:
            return next(iter(unique_partitions_defs))
        else:
            check.failed(
                "There is no PartitionsDefinition shared by all the provided assets."
                f" {len(unique_partitions_defs)} unique PartitionsDefinitions."
            )

    @cached_property
    def _sensor_mappings(
        self,
    ) -> tuple[
        Mapping[str, Sequence["RemoteSensor"]],
        Mapping[AssetKey, Sequence["RemoteSensor"]],
    ]:
        asset_key_mapping = defaultdict(list)
        job_name_mapping = defaultdict(list)
        for sensor in self.get_sensors():
            for target in sensor.get_targets():
                job_name_mapping[target.job_name].append(sensor)

            if sensor and sensor.asset_selection:
                try:
                    keys = sensor.asset_selection.resolve(self.asset_graph)
                    for key in keys:
                        # only count an asset as targeted by an automation condition sensor if it
                        # has an automation condition
                        if sensor.sensor_type in (
                            SensorType.AUTO_MATERIALIZE,
                            SensorType.AUTOMATION,
                        ):
                            node_snap = self.get_asset_node_snap(key)
                            if not node_snap or not (
                                node_snap.automation_condition
                                or node_snap.automation_condition_snapshot
                            ):
                                continue
                        asset_key_mapping[key].append(sensor)
                except DagsterError:
                    pass

        return job_name_mapping, asset_key_mapping

    @property
    def sensors_by_job_name(self) -> Mapping[str, Sequence["RemoteSensor"]]:
        return self._sensor_mappings[0]

    @property
    def _sensors_by_asset_key(self) -> Mapping[AssetKey, Sequence["RemoteSensor"]]:
        return self._sensor_mappings[1]

    @cached_property
    def schedules_by_job_name(self) -> Mapping[str, Sequence["RemoteSchedule"]]:
        mapping = defaultdict(list)
        for schedule in self.get_schedules():
            mapping[schedule.job_name].append(schedule)

        return mapping

    def get_sensors_targeting(self, asset_key: AssetKey) -> AbstractSet["RemoteSensor"]:
        asset_snap = self.get_asset_node_snap(asset_key)
        if not asset_snap:
            return _empty_set

        sensors = set()
        if asset_key in self._sensors_by_asset_key:
            sensors.update(self._sensors_by_asset_key[asset_key])

        for job_name in asset_snap.job_names:
            if job_name != IMPLICIT_ASSET_JOB_NAME and job_name in self.sensors_by_job_name:
                sensors.update(self.sensors_by_job_name[job_name])

        return sensors

    def get_schedules_targeting(self, asset_key: AssetKey) -> AbstractSet["RemoteSchedule"]:
        asset_snap = self.get_asset_node_snap(asset_key)
        if not asset_snap:
            return _empty_set

        schedules = set()
        for job_name in asset_snap.job_names:
            if job_name != IMPLICIT_ASSET_JOB_NAME and job_name in self.schedules_by_job_name:
                schedules.update(self.schedules_by_job_name[job_name])

        return schedules


class RemoteJob(RepresentedJob):
    """RemoteJob is a object that represents a loaded job definition that
    is resident in another process or container. Host processes such as dagster-webserver use
    objects such as these to interact with user-defined artifacts.
    """

    def __init__(
        self,
        job_data_snap: Optional[JobDataSnap],
        repository_handle: RepositoryHandle,
        job_ref_snap: Optional[JobRefSnap] = None,
        ref_to_data_fn: Optional[Callable[[JobRefSnap], JobDataSnap]] = None,
    ):
        check.inst_param(repository_handle, "repository_handle", RepositoryHandle)
        check.opt_inst_param(job_data_snap, "job_data", JobDataSnap)

        self._repository_handle = repository_handle

        self._memo_lock = RLock()
        self._index: Optional[JobIndex] = None

        self._job_data_snap = job_data_snap
        self._job_ref_snap = job_ref_snap
        self._ref_to_data_fn = ref_to_data_fn

        if job_data_snap:
            self._active_preset_dict = {ap.name: ap for ap in job_data_snap.active_presets}
            self._name = job_data_snap.name

        elif job_ref_snap:
            self._active_preset_dict = {ap.name: ap for ap in job_ref_snap.active_presets}
            self._name = job_ref_snap.name
            if ref_to_data_fn is None:
                check.failed("ref_to_data_fn must be passed when using deferred snapshots")

        else:
            check.failed("Expected either job data or ref, got neither")

        self._handle = JobHandle(
            job_name=self._name,
            repository_handle=repository_handle,
        )

    @property
    def _job_index(self) -> JobIndex:
        with self._memo_lock:
            if self._index is None:
                self._index = JobIndex(
                    self.job_data_snap.job,
                    self.job_data_snap.parent_job,
                )
            return self._index

    @property
    def name(self) -> str:
        return self._name

    @property
    def description(self):
        return self._job_index.job_snapshot.description

    @property
    def node_names_in_topological_order(self):
        return self._job_index.job_snapshot.node_names_in_topological_order

    @property
    def job_data_snap(self) -> JobDataSnap:
        with self._memo_lock:
            if self._job_data_snap is None:
                if self._job_ref_snap is None or self._ref_to_data_fn is None:
                    check.failed("unexpected state - unable to load data from ref")
                self._job_data_snap = self._ref_to_data_fn(self._job_ref_snap)

            return self._job_data_snap

    @property
    def repository_handle(self) -> RepositoryHandle:
        return self._repository_handle

    @property
    def op_selection(self) -> Optional[Sequence[str]]:
        return (
            self._job_index.job_snapshot.lineage_snapshot.op_selection
            if self._job_index.job_snapshot.lineage_snapshot
            else None
        )

    @property
    def resolved_op_selection(self) -> Optional[AbstractSet[str]]:
        return (
            self._job_index.job_snapshot.lineage_snapshot.resolved_op_selection
            if self._job_index.job_snapshot.lineage_snapshot
            else None
        )

    @property
    def asset_selection(self) -> Optional[AbstractSet[AssetKey]]:
        return (
            self._job_index.job_snapshot.lineage_snapshot.asset_selection
            if self._job_index.job_snapshot.lineage_snapshot
            else None
        )

    @property
    def asset_check_selection(self) -> Optional[AbstractSet[AssetCheckKey]]:
        return (
            self._job_index.job_snapshot.lineage_snapshot.asset_check_selection
            if self._job_index.job_snapshot.lineage_snapshot
            else None
        )

    @property
    def active_presets(self) -> Sequence[PresetSnap]:
        return list(self._active_preset_dict.values())

    @property
    def node_names(self) -> Sequence[str]:
        return self._job_index.job_snapshot.node_names

    def has_node_invocation(self, node_name: str):
        check.str_param(node_name, "node_name")
        return self._job_index.has_node_invocation(node_name)

    def has_preset(self, preset_name: str) -> bool:
        check.str_param(preset_name, "preset_name")
        return preset_name in self._active_preset_dict

    def get_preset(self, preset_name: str) -> PresetSnap:
        check.str_param(preset_name, "preset_name")
        return self._active_preset_dict[preset_name]

    @property
    def root_config_key(self) -> Optional[str]:
        return self.get_mode_def_snap(DEFAULT_MODE_NAME).root_config_key

    @property
    def tags(self) -> Mapping[str, str]:
        return self._job_index.job_snapshot.tags

    @property
    def run_tags(self) -> Mapping[str, str]:
        snapshot_tags = self._job_index.job_snapshot.run_tags
        # Snapshot tags will be None for snapshots originating from old code servers before the
        # introduction of run tags. In these cases, the job definition tags are treated as run tags
        # to maintain backcompat.
        return snapshot_tags if snapshot_tags is not None else self.tags

    @property
    def metadata(self) -> Mapping[str, MetadataValue]:
        return self._job_index.job_snapshot.metadata

    @property
    def job_snapshot(self) -> JobSnap:
        return self._job_index.job_snapshot

    @property
    def _snapshot_id(self) -> str:
        if self._job_ref_snap:
            return self._job_ref_snap.snapshot_id

        return self._job_index.job_snapshot_id

    @property
    def computed_job_snapshot_id(self) -> str:
        return self._snapshot_id

    @property
    def identifying_job_snapshot_id(self) -> str:
        return self._snapshot_id

    @property
    def handle(self) -> JobHandle:
        return self._handle

    def get_python_origin(self) -> JobPythonOrigin:
        repository_python_origin = self.repository_handle.get_python_origin()
        return JobPythonOrigin(self.name, repository_python_origin)

    def get_remote_origin(self) -> RemoteJobOrigin:
        return self.handle.get_remote_origin()

    def get_remote_origin_id(self) -> str:
        return self.get_remote_origin().get_id()


class RemoteExecutionPlan:
    """RemoteExecutionPlan is a object that represents an execution plan that
    was compiled in another process or persisted in an instance.
    """

    def __init__(self, execution_plan_snapshot: ExecutionPlanSnapshot):
        self.execution_plan_snapshot = check.inst_param(
            execution_plan_snapshot, "execution_plan_snapshot", ExecutionPlanSnapshot
        )

        self._step_index: Mapping[str, ExecutionStepSnap] = {
            step.key: step for step in self.execution_plan_snapshot.steps
        }

        self._step_keys_in_plan: AbstractSet[str] = (
            set(execution_plan_snapshot.step_keys_to_execute)
            if execution_plan_snapshot.step_keys_to_execute
            else set(self._step_index.keys())
        )

        self._deps = None
        self._topological_steps = None
        self._topological_step_levels = None

    @property
    def step_keys_in_plan(self) -> Sequence[str]:
        return list(self._step_keys_in_plan)

    def has_step(self, key: str) -> bool:
        check.str_param(key, "key")
        handle = StepHandle.parse_from_key(key)
        if isinstance(handle, ResolvedFromDynamicStepHandle):
            return handle.unresolved_form.to_key() in self._step_index
        return key in self._step_index

    def get_step_by_key(self, key: str):
        check.str_param(key, "key")
        return self._step_index[key]

    def get_steps_in_plan(self):
        return [self._step_index[sk] for sk in self._step_keys_in_plan]

    def key_in_plan(self, key: str):
        return key in self._step_keys_in_plan

    # Everything below this line is a near-copy of the equivalent methods on
    # ExecutionPlan. We should resolve this, probably eventually by using the
    # snapshots to support the existing ExecutionPlan methods.
    # https://github.com/dagster-io/dagster/issues/2462
    def execution_deps(self):
        if self._deps is None:
            deps = {}

            for key in self._step_keys_in_plan:
                deps[key] = set()

            for key in self._step_keys_in_plan:
                step = self._step_index[key]
                for step_input in step.inputs:
                    deps[step.key].update(
                        {
                            output_handle.step_key
                            for output_handle in step_input.upstream_output_handles
                        }.intersection(self._step_keys_in_plan)
                    )
            self._deps = deps

        return self._deps

    def topological_steps(self):
        if self._topological_steps is None:
            self._topological_steps = [
                step for step_level in self.topological_step_levels() for step in step_level
            ]

        return self._topological_steps

    def topological_step_levels(self):
        if self._topological_step_levels is None:
            self._topological_step_levels = [
                [self._step_index[step_key] for step_key in sorted(step_key_level)]
                for step_key_level in toposort(self.execution_deps())
            ]

        return self._topological_step_levels


class RemoteResource:
    """Represents a top-level resource in a repository, e.g. one passed through the Definitions API."""

    def __init__(self, resource_snap: ResourceSnap, handle: RepositoryHandle):
        self._resource_snap = check.inst_param(resource_snap, "resource_snap", ResourceSnap)
        self._handle = InstigatorHandle(
            instigator_name=self._resource_snap.name,
            repository_handle=check.inst_param(handle, "handle", RepositoryHandle),
        )

    @property
    def name(self) -> str:
        return self._resource_snap.name

    @property
    def description(self) -> Optional[str]:
        return self._resource_snap.resource_snapshot.description

    @property
    def config_field_snaps(self) -> list[ConfigFieldSnap]:
        return self._resource_snap.config_field_snaps

    @property
    def configured_values(self) -> dict[str, ResourceValueSnap]:
        return self._resource_snap.configured_values

    @property
    def config_schema_snap(self) -> ConfigSchemaSnapshot:
        return self._resource_snap.config_schema_snap

    @property
    def nested_resources(self) -> dict[str, NestedResource]:
        return self._resource_snap.nested_resources

    @property
    def parent_resources(self) -> dict[str, str]:
        return self._resource_snap.parent_resources

    @property
    def resource_type(self) -> str:
        return self._resource_snap.resource_type

    @property
    def is_top_level(self) -> bool:
        return self._resource_snap.is_top_level

    @property
    def asset_keys_using(self) -> list[AssetKey]:
        return self._resource_snap.asset_keys_using

    @property
    def job_ops_using(self) -> list[ResourceJobUsageEntry]:
        return self._resource_snap.job_ops_using

    @property
    def schedules_using(self) -> list[str]:
        return self._resource_snap.schedules_using

    @property
    def sensors_using(self) -> list[str]:
        return self._resource_snap.sensors_using

    @property
    def is_dagster_maintained(self) -> bool:
        return self._resource_snap.dagster_maintained


class RemoteSchedule:
    def __init__(self, schedule_snap: ScheduleSnap, handle: RepositoryHandle):
        self._schedule_snap = check.inst_param(schedule_snap, "schedule_snap", ScheduleSnap)
        self._handle = InstigatorHandle(
            self._schedule_snap.name,
            check.inst_param(handle, "handle", RepositoryHandle),
        )

    @property
    def name(self) -> str:
        return self._schedule_snap.name

    @property
    def cron_schedule(self) -> Union[str, Sequence[str]]:
        return self._schedule_snap.cron_schedule

    @property
    def execution_timezone(self) -> Optional[str]:
        return self._schedule_snap.execution_timezone

    @property
    def op_selection(self) -> Optional[Sequence[str]]:
        return self._schedule_snap.op_selection

    @property
    def job_name(self) -> str:
        return self._schedule_snap.job_name

    @property
    def asset_selection(self) -> Optional[AssetSelection]:
        return self._schedule_snap.asset_selection

    @property
    def mode(self) -> Optional[str]:
        return self._schedule_snap.mode

    @property
    def description(self) -> Optional[str]:
        return self._schedule_snap.description

    @property
    def partition_set_name(self) -> Optional[str]:
        return self._schedule_snap.partition_set_name

    @property
    def environment_vars(self) -> Optional[Mapping[str, str]]:
        return self._schedule_snap.environment_vars

    @property
    def handle(self) -> InstigatorHandle:
        return self._handle

    @property
    def tags(self) -> Mapping[str, str]:
        return self._schedule_snap.tags

    @property
    def metadata(self) -> Mapping[str, MetadataValue]:
        return self._schedule_snap.metadata

    def get_remote_origin(self) -> RemoteInstigatorOrigin:
        return self.handle.get_remote_origin()

    def get_remote_origin_id(self) -> str:
        return self.get_remote_origin().get_id()

    @property
    def selector(self) -> InstigatorSelector:
        return InstigatorSelector(
            location_name=self.handle.location_name,
            repository_name=self.handle.repository_name,
            name=self._schedule_snap.name,
        )

    @property
    def schedule_selector(self) -> ScheduleSelector:
        return ScheduleSelector(
            location_name=self.handle.location_name,
            repository_name=self.handle.repository_name,
            schedule_name=self._schedule_snap.name,
        )

    @cached_property
    def selector_id(self) -> str:
        return create_snapshot_id(self.selector)

    def get_compound_id(self) -> CompoundID:
        return CompoundID(
            remote_origin_id=self.get_remote_origin_id(),
            selector_id=self.selector_id,
        )

    @property
    def default_status(self) -> DefaultScheduleStatus:
        return self._schedule_snap.default_status or DefaultScheduleStatus.STOPPED

    def get_current_instigator_state(
        self, stored_state: Optional["InstigatorState"]
    ) -> "InstigatorState":
        from dagster._core.scheduler.instigation import (
            InstigatorState,
            InstigatorStatus,
            ScheduleInstigatorData,
        )

        if self.default_status == DefaultScheduleStatus.RUNNING:
            if stored_state:
                return stored_state

            return InstigatorState(
                self.get_remote_origin(),
                InstigatorType.SCHEDULE,
                InstigatorStatus.DECLARED_IN_CODE,
                ScheduleInstigatorData(self.cron_schedule, start_timestamp=None),
            )
        else:
            # Ignore DECLARED_IN_CODE states in the DB if the default status
            # isn't DefaultScheduleStatus.RUNNING - this would indicate that the schedule's
            # default has been changed in code but there's still a lingering DECLARED_IN_CODE
            # row in the database that can be ignored
            if stored_state:
                return (
                    stored_state.with_status(InstigatorStatus.STOPPED)
                    if stored_state.status == InstigatorStatus.DECLARED_IN_CODE
                    else stored_state
                )

            return InstigatorState(
                self.get_remote_origin(),
                InstigatorType.SCHEDULE,
                InstigatorStatus.STOPPED,
                ScheduleInstigatorData(self.cron_schedule, start_timestamp=None),
            )

    def execution_time_iterator(
        self, start_timestamp: float, ascending: bool = True
    ) -> Iterator[datetime]:
        return schedule_execution_time_iterator(
            start_timestamp, self.cron_schedule, self.execution_timezone, ascending
        )


class RemoteSensor:
    def __init__(self, sensor_snap: SensorSnap, handle: RepositoryHandle):
        self._sensor_snap = check.inst_param(sensor_snap, "sensor_snap", SensorSnap)
        self._handle = InstigatorHandle(
            self._sensor_snap.name, check.inst_param(handle, "handle", RepositoryHandle)
        )

    @property
    def name(self) -> str:
        return self._sensor_snap.name

    @property
    def handle(self) -> InstigatorHandle:
        return self._handle

    @property
    def job_name(self) -> Optional[str]:
        target = self._get_single_target()
        return target.job_name if target else None

    @property
    def asset_selection(self) -> Optional[AssetSelection]:
        return self._sensor_snap.asset_selection

    @property
    def mode(self) -> Optional[str]:
        target = self._get_single_target()
        return target.mode if target else None

    @property
    def op_selection(self) -> Optional[Sequence[str]]:
        target = self._get_single_target()
        return target.op_selection if target else None

    def _get_single_target(self) -> Optional[TargetSnap]:
        if self._sensor_snap.target_dict:
            return next(iter(self._sensor_snap.target_dict.values()))
        else:
            return None

    def get_target(self, job_name: Optional[str] = None) -> Optional[TargetSnap]:
        if job_name:
            return self._sensor_snap.target_dict[job_name]
        else:
            return self._get_single_target()

    def get_targets(self) -> Sequence[TargetSnap]:
        return list(self._sensor_snap.target_dict.values())

    @property
    def description(self) -> Optional[str]:
        return self._sensor_snap.description

    @property
    def min_interval_seconds(self) -> int:
        if isinstance(self._sensor_snap, SensorSnap) and self._sensor_snap.min_interval:
            return self._sensor_snap.min_interval
        return DEFAULT_SENSOR_DAEMON_INTERVAL

    @property
    def run_tags(self) -> Mapping[str, str]:
        return self._sensor_snap.run_tags

    def get_remote_origin(self) -> RemoteInstigatorOrigin:
        return self._handle.get_remote_origin()

    def get_remote_origin_id(self) -> str:
        return self.get_remote_origin().get_id()

    @property
    def selector(self) -> InstigatorSelector:
        return InstigatorSelector(
            location_name=self.handle.location_name,
            repository_name=self.handle.repository_name,
            name=self._sensor_snap.name,
        )

    @property
    def sensor_selector(self) -> SensorSelector:
        return SensorSelector(
            location_name=self.handle.location_name,
            repository_name=self.handle.repository_name,
            sensor_name=self._sensor_snap.name,
        )

    @cached_property
    def selector_id(self) -> str:
        return create_snapshot_id(self.selector)

    def get_compound_id(self) -> CompoundID:
        return CompoundID(
            remote_origin_id=self.get_remote_origin_id(),
            selector_id=self.selector_id,
        )

    @property
    def sensor_type(self) -> SensorType:
        return self._sensor_snap.sensor_type or SensorType.UNKNOWN

    def get_current_instigator_state(
        self, stored_state: Optional["InstigatorState"]
    ) -> "InstigatorState":
        from dagster._core.scheduler.instigation import (
            InstigatorState,
            InstigatorStatus,
            SensorInstigatorData,
        )

        if self.default_status == DefaultSensorStatus.RUNNING:
            return (
                stored_state
                if stored_state
                else InstigatorState(
                    self.get_remote_origin(),
                    InstigatorType.SENSOR,
                    InstigatorStatus.DECLARED_IN_CODE,
                    SensorInstigatorData(
                        min_interval=self.min_interval_seconds,
                        sensor_type=self.sensor_type,
                    ),
                )
            )
        else:
            # Ignore DECLARED_IN_CODE states in the DB if the default status
            # isn't DefaultSensorStatus.RUNNING - this would indicate that the schedule's
            # default has changed
            if stored_state:
                return (
                    stored_state.with_status(InstigatorStatus.STOPPED)
                    if stored_state.status == InstigatorStatus.DECLARED_IN_CODE
                    else stored_state
                )

            return InstigatorState(
                self.get_remote_origin(),
                InstigatorType.SENSOR,
                InstigatorStatus.STOPPED,
                SensorInstigatorData(
                    min_interval=self.min_interval_seconds,
                    sensor_type=self.sensor_type,
                ),
            )

    @property
    def metadata(self) -> Optional[SensorMetadataSnap]:
        return self._sensor_snap.metadata

    @property
    def tags(self) -> Mapping[str, str]:
        return self._sensor_snap.tags

    @property
    def default_status(self) -> DefaultSensorStatus:
        return self._sensor_snap.default_status or DefaultSensorStatus.STOPPED


class RemotePartitionSet:
    def __init__(self, partition_set_snap: PartitionSetSnap, handle: RepositoryHandle):
        self._partition_set_snap = check.inst_param(
            partition_set_snap, "partition_set_snap", PartitionSetSnap
        )
        self._handle = PartitionSetHandle(
            partition_set_name=partition_set_snap.name,
            repository_handle=check.inst_param(handle, "handle", RepositoryHandle),
        )

    @property
    def name(self) -> str:
        return self._partition_set_snap.name

    @property
    def op_selection(self) -> Optional[Sequence[str]]:
        return self._partition_set_snap.op_selection

    @property
    def mode(self) -> Optional[str]:
        return self._partition_set_snap.mode

    @property
    def job_name(self) -> str:
        return self._partition_set_snap.job_name

    @property
    def backfill_policy(self) -> Optional[BackfillPolicy]:
        return self._partition_set_snap.backfill_policy

    @property
    def repository_handle(self) -> RepositoryHandle:
        return self._handle.repository_handle

    def get_remote_origin(self) -> RemotePartitionSetOrigin:
        return self._handle.get_remote_origin()

    def get_remote_origin_id(self) -> str:
        return self.get_remote_origin().get_id()

    def has_partition_name_data(self) -> bool:
        # Partition sets from older versions of Dagster as well as partition sets using
        # a DynamicPartitionsDefinition require calling out to user code to compute the partition
        # names
        return self._partition_set_snap.partitions is not None

    def has_partitions_definition(self) -> bool:
        # Partition sets from older versions of Dagster as well as partition sets using
        # a DynamicPartitionsDefinition require calling out to user code to get the
        # partitions definition
        return self._partition_set_snap.partitions is not None

    def get_partitions_definition(self) -> PartitionsDefinition:
        partitions_data = self._partition_set_snap.partitions
        if partitions_data is None:
            check.failed(
                "Partition set does not have partition data, cannot get partitions definition"
            )
        return partitions_data.get_partitions_definition()

    def get_partition_names(self, instance: DagsterInstance) -> Sequence[str]:
        partitions = self._partition_set_snap.partitions
        if partitions is None:
            check.failed(
                "Partition set does not have partition data, cannot get partitions definition"
            )
        return self.get_partitions_definition().get_partition_keys(
            dynamic_partitions_store=instance
        )



import pkgutil
from collections.abc import Iterable, Iterator, Mapping
from importlib import import_module
from types import ModuleType
from typing import Union

from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.asset_spec import AssetSpec
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.source_asset import SourceAsset


def find_objects_in_module_of_types(
    module: ModuleType,
    types: Union[type, tuple[type, ...]],
) -> Iterator:
    """Yields instances or subclasses of the given type(s)."""
    for attr in dir(module):
        value = getattr(module, attr)
        if isinstance(value, types):
            yield value
        elif isinstance(value, list) and all(isinstance(el, types) for el in value):
            yield from value


def find_subclasses_in_module(
    module: ModuleType,
    types: Union[type, tuple[type, ...]],
) -> Iterator:
    """Yields instances or subclasses of the given type(s)."""
    for attr in dir(module):
        value = getattr(module, attr)
        if isinstance(value, type) and issubclass(value, types):
            yield value


def key_iterator(
    asset: Union[AssetsDefinition, SourceAsset, AssetSpec], included_targeted_keys: bool = False
) -> Iterator[AssetKey]:
    return (
        iter(
            [
                *asset.keys,
                *(
                    [check_key.asset_key for check_key in asset.check_keys]
                    if included_targeted_keys
                    else []
                ),
            ]
        )
        if isinstance(asset, AssetsDefinition)
        else iter([asset.key])
    )


def find_modules_in_package(package_module: ModuleType) -> Iterable[ModuleType]:
    yield package_module
    if package_module.__file__:
        for _, modname, is_pkg in pkgutil.walk_packages(
            package_module.__path__, prefix=package_module.__name__ + "."
        ):
            submodule = import_module(modname)
            if is_pkg:
                yield from find_modules_in_package(submodule)
            else:
                yield submodule
    else:
        raise ValueError(
            f"Tried to find modules in package {package_module}, but its __file__ is None"
        )


def replace_keys_in_asset(
    asset: Union[AssetsDefinition, AssetSpec, SourceAsset],
    key_replacements: Mapping[AssetKey, AssetKey],
) -> Union[AssetsDefinition, AssetSpec, SourceAsset]:
    if isinstance(asset, SourceAsset):
        return asset.with_attributes(key=key_replacements.get(asset.key, asset.key))
    if isinstance(asset, AssetSpec):
        return asset.replace_attributes(
            key=key_replacements.get(asset.key, asset.key),
        )
    else:
        updated_object = asset.with_attributes(asset_key_replacements=key_replacements)
        return updated_object



import datetime
from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Generic, Optional, Union

from typing_extensions import Self

import dagster._check as check
from dagster._annotations import beta, public
from dagster._core.asset_graph_view.entity_subset import EntitySubset
from dagster._core.asset_graph_view.serializable_entity_subset import SerializableEntitySubset
from dagster._core.definitions.asset_key import (
    AssetCheckKey,
    AssetKey,
    CoercibleToAssetKey,
    T_EntityKey,
)
from dagster._core.definitions.declarative_automation.serialized_objects import (
    AssetSubsetWithMetadata,
    AutomationConditionCursor,
    AutomationConditionEvaluation,
    AutomationConditionNodeCursor,
    AutomationConditionNodeSnapshot,
    AutomationConditionSnapshot,
    get_serializable_candidate_subset,
)
from dagster._core.definitions.partition import AllPartitionsSubset
from dagster._core.definitions.time_window_partitions import TimeWindowPartitionsSubset
from dagster._record import copy, record
from dagster._serdes.serdes import is_whitelisted_for_serdes_object
from dagster._time import get_current_timestamp
from dagster._utils.security import non_secure_md5_hash_str
from dagster._utils.warnings import disable_dagster_warnings

if TYPE_CHECKING:
    from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy
    from dagster._core.definitions.declarative_automation.automation_context import (
        AutomationContext,
    )
    from dagster._core.definitions.declarative_automation.operators import AndAutomationCondition
    from dagster._core.definitions.declarative_automation.operators.dep_operators import (
        DepsAutomationCondition,
    )


class AutomationCondition(ABC, Generic[T_EntityKey]):
    """An AutomationCondition represents a condition of an asset that impacts whether it should be
    automatically executed. For example, you can have a condition which becomes true whenever the
    code version of the asset is changed, or whenever an upstream dependency is updated.

    .. code-block:: python

        from dagster import AutomationCondition, asset

        @asset(automation_condition=AutomationCondition.on_cron("0 0 * * *"))
        def my_asset(): ...

    AutomationConditions may be combined together into expressions using a variety of operators.

    .. code-block:: python

        from dagster import AssetSelection, AutomationCondition, asset

        # any dependencies from the "important" group are missing
        any_important_deps_missing = AutomationCondition.any_deps_match(
            AutomationCondition.missing(),
        ).allow(AssetSelection.groups("important"))

        # there is a new code version for this asset since the last time it was requested
        new_code_version = AutomationCondition.code_version_changed().since(
            AutomationCondition.newly_requested()
        )

        # there is a new code version and no important dependencies are missing
        my_condition = new_code_version & ~any_important_deps_missing

        @asset(automation_condition=my_condition)
        def my_asset(): ...

    """

    @property
    def requires_cursor(self) -> bool:
        return True

    @property
    def children(self) -> Sequence["AutomationCondition"]:
        return []

    @property
    def description(self) -> str:
        """Human-readable description of when this condition is true."""
        return ""

    @property
    def name(self) -> str:
        """Formal name of this specific condition, generally aligning with its static constructor."""
        return self.__class__.__name__

    def get_label(self) -> Optional[str]:
        return None

    def get_node_snapshot(self, unique_id: str) -> AutomationConditionNodeSnapshot:
        """Returns a snapshot of this condition that can be used for serialization."""
        return AutomationConditionNodeSnapshot(
            class_name=self.__class__.__name__,
            description=self.description,
            unique_id=unique_id,
            label=self.get_label(),
            name=self.name,
        )

    def get_snapshot(
        self, *, parent_unique_id: Optional[str] = None, index: Optional[int] = None
    ) -> AutomationConditionSnapshot:
        """Returns a serializable snapshot of the entire AutomationCondition tree."""
        unique_id = self.get_node_unique_id(parent_unique_id=parent_unique_id, index=index)
        node_snapshot = self.get_node_snapshot(unique_id)
        children = [
            child.get_snapshot(parent_unique_id=unique_id, index=i)
            for (i, child) in enumerate(self.children)
        ]
        return AutomationConditionSnapshot(node_snapshot=node_snapshot, children=children)

    def get_node_unique_id(self, *, parent_unique_id: Optional[str], index: Optional[int]) -> str:
        """Returns a unique identifier for this condition within the broader condition tree."""
        parts = [str(parent_unique_id), str(index), self.name]
        return non_secure_md5_hash_str("".join(parts).encode())

    def get_unique_id(
        self, *, parent_node_unique_id: Optional[str] = None, index: Optional[int] = None
    ) -> str:
        """Returns a unique identifier for the entire subtree."""
        node_unique_id = self.get_node_unique_id(
            parent_unique_id=parent_node_unique_id, index=index
        )
        child_unique_ids = [
            child.get_unique_id(parent_node_unique_id=node_unique_id, index=i)
            for i, child in enumerate(self.children)
        ]
        return non_secure_md5_hash_str("".join([node_unique_id, *child_unique_ids]).encode())

    def __hash__(self) -> int:
        return hash(self.get_unique_id())

    @property
    def has_rule_condition(self) -> bool:
        from dagster._core.definitions.declarative_automation.legacy import RuleCondition

        if isinstance(self, RuleCondition):
            return True
        return any(child.has_rule_condition for child in self.children)

    @property
    def is_serializable(self) -> bool:
        if not is_whitelisted_for_serdes_object(self):
            return False
        return all(child.is_serializable for child in self.children)

    def as_auto_materialize_policy(self) -> "AutoMaterializePolicy":
        """Returns an AutoMaterializePolicy which contains this condition."""
        from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy

        with disable_dagster_warnings():
            return AutoMaterializePolicy.from_automation_condition(self)

    @abstractmethod
    def evaluate(
        self, context: "AutomationContext[T_EntityKey]"
    ) -> "AutomationResult[T_EntityKey]":
        raise NotImplementedError()

    def __and__(
        self, other: "AutomationCondition[T_EntityKey]"
    ) -> "AndAutomationCondition[T_EntityKey]":
        from dagster._core.definitions.declarative_automation.operators import (
            AndAutomationCondition,
        )

        # Consolidate any unlabeled `AndAutomationCondition`s together.
        return AndAutomationCondition(
            operands=[
                *(
                    self.operands
                    if isinstance(self, AndAutomationCondition) and self.label is None
                    else (self,)
                ),
                *(
                    other.operands
                    if isinstance(other, AndAutomationCondition) and other.label is None
                    else (other,)
                ),
            ]
        )

    def __or__(
        self, other: "AutomationCondition[T_EntityKey]"
    ) -> "BuiltinAutomationCondition[T_EntityKey]":
        from dagster._core.definitions.declarative_automation.operators import OrAutomationCondition

        # Consolidate any unlabeled `OrAutomationCondition`s together.
        return OrAutomationCondition(
            operands=[
                *(
                    self.operands
                    if isinstance(self, OrAutomationCondition) and self.label is None
                    else (self,)
                ),
                *(
                    other.operands
                    if isinstance(other, OrAutomationCondition) and other.label is None
                    else (other,)
                ),
            ]
        )

    def __invert__(self) -> "BuiltinAutomationCondition[T_EntityKey]":
        from dagster._core.definitions.declarative_automation.operators import (
            NotAutomationCondition,
        )

        return NotAutomationCondition(operand=self)

    def since(
        self, reset_condition: "AutomationCondition[T_EntityKey]"
    ) -> "BuiltinAutomationCondition[T_EntityKey]":
        """Returns an AutomationCondition that is true if this condition has become true since the
        last time the reference condition became true.
        """
        from dagster._core.definitions.declarative_automation.operators import SinceCondition

        return SinceCondition(trigger_condition=self, reset_condition=reset_condition)

    def newly_true(self) -> "BuiltinAutomationCondition[T_EntityKey]":
        """Returns an AutomationCondition that is true only on the tick that this condition goes
        from false to true for a given target.
        """
        from dagster._core.definitions.declarative_automation.operators import NewlyTrueCondition

        return NewlyTrueCondition(operand=self)

    def since_last_handled(self: "AutomationCondition") -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition that is true if this condition has become true since the
        target was last requested or updated, and since the last time this target's condition was modified.
        """
        with disable_dagster_warnings():
            return self.since(
                (
                    AutomationCondition.newly_requested()
                    | AutomationCondition.newly_updated()
                    | AutomationCondition.initial_evaluation()
                ).with_label("handled")
            )

    @public
    def replace(
        self, old: Union["AutomationCondition", str], new: "AutomationCondition"
    ) -> "AutomationCondition":
        """Replaces all instances of ``old`` across any sub-conditions with ``new``.

        If ``old`` is a string, then conditions with a label matching
        that string will be replaced.

        Args:
            old (Union[AutomationCondition, str]): The condition to replace.
            new (AutomationCondition): The condition to replace with.
        """
        return new if old in [self, self.get_label()] else self

    @public
    @staticmethod
    def asset_matches(
        key: "CoercibleToAssetKey", condition: "AutomationCondition[AssetKey]"
    ) -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition that is true if this condition is true for the given key."""
        from dagster._core.definitions.declarative_automation.operators import (
            EntityMatchesCondition,
        )

        asset_key = AssetKey.from_coercible(key)
        return EntityMatchesCondition(key=asset_key, operand=condition)

    @public
    @staticmethod
    def any_deps_match(condition: "AutomationCondition") -> "DepsAutomationCondition":
        """Returns an AutomationCondition that is true for a if at least one partition
        of the any of the target's dependencies evaluate to True for the given condition.

        Args:
            condition (AutomationCondition): The AutomationCondition that will be evaluated against
                this target's dependencies.
        """
        from dagster._core.definitions.declarative_automation.operators import AnyDepsCondition

        return AnyDepsCondition(operand=condition)

    @public
    @staticmethod
    def all_deps_match(condition: "AutomationCondition") -> "DepsAutomationCondition":
        """Returns an AutomationCondition that is true for a if at least one partition
        of the all of the target's dependencies evaluate to True for the given condition.

        Args:
            condition (AutomationCondition): The AutomationCondition that will be evaluated against
                this target's dependencies.
        """
        from dagster._core.definitions.declarative_automation.operators import AllDepsCondition

        return AllDepsCondition(operand=condition)

    @public
    @staticmethod
    def any_checks_match(
        condition: "AutomationCondition[AssetCheckKey]", blocking_only: bool = False
    ) -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition that is true for if at least one of the target's
        checks evaluate to True for the given condition.

        Args:
            condition (AutomationCondition): The AutomationCondition that will be evaluated against
                this asset's checks.
            blocking_only (bool): Determines if this condition will only be evaluated against blocking
                checks. Defaults to False.
        """
        from dagster._core.definitions.declarative_automation.operators import AnyChecksCondition

        return AnyChecksCondition(operand=condition, blocking_only=blocking_only)

    @public
    @staticmethod
    def all_checks_match(
        condition: "AutomationCondition[AssetCheckKey]", blocking_only: bool = False
    ) -> "BuiltinAutomationCondition[AssetKey]":
        """Returns an AutomationCondition that is true for an asset partition if all of its checks
        evaluate to True for the given condition.

        Args:
            condition (AutomationCondition): The AutomationCondition that will be evaluated against
                this asset's checks.
            blocking_only (bool): Determines if this condition will only be evaluated against blocking
                checks. Defaults to False.
        """
        from dagster._core.definitions.declarative_automation.operators import AllChecksCondition

        return AllChecksCondition(operand=condition, blocking_only=blocking_only)

    @public
    @staticmethod
    def missing() -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition that is true if the target has not been executed."""
        from dagster._core.definitions.declarative_automation.operands import (
            MissingAutomationCondition,
        )

        return MissingAutomationCondition()

    @public
    @staticmethod
    def run_in_progress() -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition that is true if the target is part of an in-progress run."""
        from dagster._core.definitions.declarative_automation.operands import (
            RunInProgressAutomationCondition,
        )

        return RunInProgressAutomationCondition()

    @public
    @staticmethod
    def backfill_in_progress() -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition that is true if the target is part of an in-progress backfill."""
        from dagster._core.definitions.declarative_automation.operands import (
            BackfillInProgressAutomationCondition,
        )

        return BackfillInProgressAutomationCondition()

    @public
    @staticmethod
    def execution_failed() -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition that is true if the latest execution of the target failed."""
        from dagster._core.definitions.declarative_automation.operands import (
            ExecutionFailedAutomationCondition,
        )

        return ExecutionFailedAutomationCondition()

    @public
    @staticmethod
    def in_progress() -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition that is true for an asset partition if it is part of an
        in-progress run or backfill.
        """
        return (
            AutomationCondition.run_in_progress() | AutomationCondition.backfill_in_progress()
        ).with_label("in_progress")

    @public
    @staticmethod
    def check_passed() -> "BuiltinAutomationCondition[AssetCheckKey]":
        """Returns an AutomationCondition that is true for an asset check if it has evaluated against
        the latest materialization of an asset and passed.
        """
        from dagster._core.definitions.declarative_automation.operands import CheckResultCondition

        return CheckResultCondition(passed=True)

    @public
    @staticmethod
    def check_failed() -> "BuiltinAutomationCondition[AssetCheckKey]":
        """Returns an AutomationCondition that is true for an asset check if it has evaluated against
        the latest materialization of an asset and failed.
        """
        from dagster._core.definitions.declarative_automation.operands import CheckResultCondition

        return CheckResultCondition(passed=False)

    @public
    @staticmethod
    def initial_evaluation() -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition that is true on the first evaluation of the expression."""
        from dagster._core.definitions.declarative_automation.operands import (
            InitialEvaluationCondition,
        )

        return InitialEvaluationCondition()

    @public
    @staticmethod
    def in_latest_time_window(
        lookback_delta: Optional[datetime.timedelta] = None,
    ) -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition that is true when the target it is within the latest
        time window.

        Args:
            lookback_delta (Optional, datetime.timedelta): If provided, the condition will
                return all partitions within the provided delta of the end of the latest time window.
                For example, if this is used on a daily-partitioned asset with a lookback_delta of
                48 hours, this will return the latest two partitions.
        """
        from dagster._core.definitions.declarative_automation.operands import (
            InLatestTimeWindowCondition,
        )

        return InLatestTimeWindowCondition.from_lookback_delta(lookback_delta)

    @public
    @staticmethod
    def will_be_requested() -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition that is true if the target will be requested this tick."""
        from dagster._core.definitions.declarative_automation.operands import (
            WillBeRequestedCondition,
        )

        return WillBeRequestedCondition()

    @public
    @staticmethod
    def newly_updated() -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition that is true if the target has been updated since the previous tick."""
        from dagster._core.definitions.declarative_automation.operands import NewlyUpdatedCondition

        return NewlyUpdatedCondition()

    @staticmethod
    def executed_with_root_target() -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition that is true if the latest run that updated the target also executed
        with the root key that the global condition is applied to.
        """
        from dagster._core.definitions.declarative_automation.operands import (
            LatestRunExecutedWithRootTargetCondition,
        )

        return LatestRunExecutedWithRootTargetCondition()

    @public
    @staticmethod
    def newly_requested() -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition that is true if the target was requested on the previous tick."""
        from dagster._core.definitions.declarative_automation.operands import (
            NewlyRequestedCondition,
        )

        return NewlyRequestedCondition()

    @public
    @staticmethod
    def code_version_changed() -> "BuiltinAutomationCondition[AssetKey]":
        """Returns an AutomationCondition that is true if the target's code version has been changed
        since the previous tick.
        """
        from dagster._core.definitions.declarative_automation.operands.operands import (
            CodeVersionChangedCondition,
        )

        return CodeVersionChangedCondition()

    @public
    @staticmethod
    def data_version_changed() -> "BuiltinAutomationCondition[AssetKey]":
        """Returns an AutomationCondition that is true if the target's data version has been changed
        since the previous tick.
        """
        from dagster._core.definitions.declarative_automation.operands.operands import (
            DataVersionChangedCondition,
        )

        return DataVersionChangedCondition()

    @public
    @staticmethod
    def cron_tick_passed(
        cron_schedule: str, cron_timezone: str = "UTC"
    ) -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition that is whenever a cron tick of the provided schedule is passed."""
        from dagster._core.definitions.declarative_automation.operands import (
            CronTickPassedCondition,
        )

        return CronTickPassedCondition(cron_schedule=cron_schedule, cron_timezone=cron_timezone)

    @public
    @staticmethod
    def newly_missing() -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition that is true on the tick that the target becomes missing."""
        return AutomationCondition.missing().newly_true().with_label("newly_missing")

    @public
    @staticmethod
    def any_deps_updated() -> "DepsAutomationCondition":
        """Returns an AutomationCondition that is true if the target has at least one dependency
        that has updated since the previous tick, or will be requested on this tick.

        Will ignore parent updates if the run that updated the parent also plans to update
        the asset or check that this condition is applied to.
        """
        return AutomationCondition.any_deps_match(
            (
                AutomationCondition.newly_updated()
                # executed_with_root_target is fairly expensive on a per-partition basis,
                # but newly_updated is bounded in the number of partitions that might be
                # updated on a single tick
                & ~AutomationCondition.executed_with_root_target()
            ).with_label("newly_updated_without_root")
            | AutomationCondition.will_be_requested()
        ).with_label("any_deps_updated")

    @public
    @staticmethod
    def any_deps_missing() -> "DepsAutomationCondition":
        """Returns an AutomationCondition that is true if the target has at least one dependency
        that is missing, and will not be requested on this tick.
        """
        return AutomationCondition.any_deps_match(
            AutomationCondition.missing() & ~AutomationCondition.will_be_requested()
        ).with_label("any_deps_missing")

    @public
    @staticmethod
    def any_deps_in_progress() -> "DepsAutomationCondition":
        """Returns an AutomationCondition that is true if the target has at least one dependency
        that is in progress.
        """
        return AutomationCondition.any_deps_match(AutomationCondition.in_progress()).with_label(
            "any_deps_in_progress"
        )

    @public
    @staticmethod
    def all_deps_blocking_checks_passed() -> "DepsAutomationCondition":
        """Returns an AutomationCondition that is true for any partition where all upstream
        blocking checks have passed, or will be requested on this tick.

        In-tick requests are allowed to enable creating runs that target both a parent with
        blocking checks and a child. Even though the checks have not currently passed, if
        they fail within the run, the run machinery will prevent the child from being
        materialized.
        """
        return AutomationCondition.all_deps_match(
            AutomationCondition.all_checks_match(
                AutomationCondition.check_passed() | AutomationCondition.will_be_requested(),
                blocking_only=True,
            ).with_label("all_blocking_checks_passed")
        ).with_label("all_deps_blocking_checks_passed")

    @public
    @staticmethod
    def all_deps_updated_since_cron(
        cron_schedule: str, cron_timezone: str = "UTC"
    ) -> "DepsAutomationCondition":
        """Returns an AutomatonCondition that is true if all of the target's dependencies have
        updated since the latest tick of the provided cron schedule.
        """
        return AutomationCondition.all_deps_match(
            AutomationCondition.newly_updated().since(
                AutomationCondition.cron_tick_passed(cron_schedule, cron_timezone)
            )
            | AutomationCondition.will_be_requested()
        ).with_label(f"all_deps_updated_since_cron({cron_schedule}, {cron_timezone})")

    @public
    @staticmethod
    def eager() -> "AndAutomationCondition":
        """Returns an AutomationCondition which will cause a target to be executed if any of
        its dependencies update, and will execute missing partitions if they become missing
        after this condition is applied to the target.

        This will not execute targets that have any missing or in progress dependencies, or
        are currently in progress.

        For time partitioned assets, only the latest time partition will be considered.
        """
        return (
            AutomationCondition.in_latest_time_window()
            & (
                AutomationCondition.newly_missing() | AutomationCondition.any_deps_updated()
            ).since_last_handled()
            & ~AutomationCondition.any_deps_missing()
            & ~AutomationCondition.any_deps_in_progress()
            & ~AutomationCondition.in_progress()
        ).with_label("eager")

    @public
    @staticmethod
    def on_cron(cron_schedule: str, cron_timezone: str = "UTC") -> "AndAutomationCondition":
        """Returns an AutomationCondition which will cause a target to be executed on a given
        cron schedule, after all of its dependencies have been updated since the latest
        tick of that cron schedule.

        For time partitioned assets, only the latest time partition will be considered.
        """
        return (
            AutomationCondition.in_latest_time_window()
            & AutomationCondition.cron_tick_passed(
                cron_schedule, cron_timezone
            ).since_last_handled()
            & AutomationCondition.all_deps_updated_since_cron(cron_schedule, cron_timezone)
        ).with_label(f"on_cron({cron_schedule}, {cron_timezone})")

    @public
    @staticmethod
    def on_missing() -> "AndAutomationCondition":
        """Returns an AutomationCondition which will execute partitions of the target that
        are added after this condition is applied to the asset.

        This will not execute targets that have any missing dependencies.

        For time partitioned assets, only the latest time partition will be considered.
        """
        return (
            AutomationCondition.in_latest_time_window()
            & (
                AutomationCondition.missing()
                .newly_true()
                .since_last_handled()
                .with_label("missing_since_last_handled")
            )
            & ~AutomationCondition.any_deps_missing()
        ).with_label("on_missing")

    @public
    @beta
    @staticmethod
    def any_downstream_conditions() -> "BuiltinAutomationCondition":
        """Returns an AutomationCondition which represents the union of all distinct downstream conditions."""
        from dagster._core.definitions.declarative_automation.operators import (
            AnyDownstreamConditionsCondition,
        )

        return AnyDownstreamConditionsCondition()


@record
class BuiltinAutomationCondition(AutomationCondition[T_EntityKey]):
    """Base class for AutomationConditions provided by the core dagster framework."""

    label: Optional[str] = None

    def get_label(self) -> Optional[str]:
        return self.label

    @public
    def with_label(self, label: Optional[str]) -> Self:
        """Returns a copy of this AutomationCondition with a human-readable label."""
        return copy(self, label=label)


class AutomationResult(Generic[T_EntityKey]):
    """The result of evaluating an AutomationCondition."""

    def __init__(
        self,
        context: "AutomationContext",
        true_subset: EntitySubset[T_EntityKey],
        cursor: Optional[str] = None,
        child_results: Optional[Sequence["AutomationResult"]] = None,
        **kwargs,
    ):
        from dagster._core.definitions.declarative_automation.automation_context import (
            AutomationContext,
        )

        self._context = check.inst_param(context, "context", AutomationContext)
        self._true_subset = check.inst_param(true_subset, "true_subset", EntitySubset)
        self._child_results = check.opt_sequence_param(
            child_results, "child_results", of_type=AutomationResult
        )

        self._start_timestamp = context.create_time.timestamp()
        self._end_timestamp = get_current_timestamp()

        # hidden_param which should only be set by legacy RuleConditions
        self._subsets_with_metadata = check.opt_sequence_param(
            kwargs.get("subsets_with_metadata"), "subsets_with_metadata", AssetSubsetWithMetadata
        )

        # hidden_param which should only be set by builtin conditions which require high performance
        # in their serdes layer
        structured_cursor = kwargs.get("structured_cursor")
        invalid_hidden_params = set(kwargs.keys()) - {"subsets_with_metadata", "structured_cursor"}
        check.param_invariant(
            not invalid_hidden_params, "kwargs", f"Invalid hidden params: {invalid_hidden_params}"
        )
        check.param_invariant(
            not (cursor and structured_cursor),
            "structured_cursor",
            "Cannot provide both cursor and structured_cursor.",
        )
        self._extra_state = check.opt_str_param(cursor, "cursor") or structured_cursor

        # used to enable the evaluator class to modify the evaluation in some edge cases
        self._serializable_subset_override: Optional[SerializableEntitySubset] = None

    @property
    def key(self) -> T_EntityKey:
        return self._true_subset.key

    @property
    def true_subset(self) -> EntitySubset[T_EntityKey]:
        return self._true_subset

    @property
    def start_timestamp(self) -> float:
        return self._start_timestamp

    @property
    def end_timestamp(self) -> float:
        return self._end_timestamp

    @property
    def child_results(self) -> Sequence["AutomationResult"]:
        return self._child_results

    @property
    def condition(self) -> AutomationCondition:
        return self._context.condition

    @property
    def condition_unique_id(self) -> str:
        return self._context.condition_unique_id

    @cached_property
    def value_hash(self) -> str:
        """An identifier for the contents of this AutomationResult. This will be identical for
        results with identical values, allowing us to avoid storing redundant information.
        """
        components: Sequence[str] = [
            self.condition_unique_id,
            self.condition.description,
            _compute_subset_value_str(self.get_serializable_subset()),
            _compute_subset_value_str(
                self._context.candidate_subset.convert_to_serializable_subset()
            ),
            *(_compute_subset_with_metadata_value_str(swm) for swm in self._subsets_with_metadata),
            *(child_result.value_hash for child_result in self._child_results),
        ]
        return non_secure_md5_hash_str("".join(components).encode("utf-8"))

    @cached_property
    def node_cursor(self) -> Optional[AutomationConditionNodeCursor]:
        """Cursor value storing information about this specific evaluation node, if required."""
        if not self.condition.requires_cursor:
            return None
        return AutomationConditionNodeCursor(
            true_subset=self.get_serializable_subset(),
            candidate_subset=get_serializable_candidate_subset(
                self._context.candidate_subset.convert_to_serializable_subset()
            ),
            subsets_with_metadata=self._subsets_with_metadata,
            extra_state=self._extra_state,
        )

    @cached_property
    def serializable_evaluation(self) -> AutomationConditionEvaluation:
        return AutomationConditionEvaluation(
            condition_snapshot=self.condition.get_node_snapshot(self.condition_unique_id),
            true_subset=self.get_serializable_subset(),
            candidate_subset=get_serializable_candidate_subset(
                self._context.candidate_subset.convert_to_serializable_subset()
            ),
            subsets_with_metadata=self._subsets_with_metadata,
            start_timestamp=self._start_timestamp,
            end_timestamp=self._end_timestamp,
            child_evaluations=[
                child_result.serializable_evaluation for child_result in self._child_results
            ],
        )

    def set_internal_serializable_subset_override(self, override: SerializableEntitySubset) -> None:
        """Internal method for handling edge cases in which the serializable evaluation must be
        updated after evaluation completes.
        """
        self._serializable_subset_override = override

    def get_child_node_cursors(self) -> Mapping[str, AutomationConditionNodeCursor]:
        node_cursors = {self.condition_unique_id: self.node_cursor} if self.node_cursor else {}
        for child_result in self._child_results:
            node_cursors.update(child_result.get_child_node_cursors())
        return node_cursors

    def get_new_cursor(self) -> AutomationConditionCursor:
        return AutomationConditionCursor(
            previous_requested_subset=self.serializable_evaluation.true_subset,
            effective_timestamp=self._context.evaluation_time.timestamp(),
            last_event_id=self._context.max_storage_id,
            node_cursors_by_unique_id=self.get_child_node_cursors(),
            result_value_hash=self.value_hash,
        )

    def get_serializable_subset(self) -> SerializableEntitySubset:
        return (
            self.true_subset.convert_to_serializable_subset() or self._serializable_subset_override
        )

    def compute_legacy_expected_data_time(self) -> Optional[datetime.datetime]:
        from dagster._core.definitions.freshness_based_auto_materialize import (
            get_expected_data_time_for_asset_key,
        )

        legacy_context = self._context._legacy_context  # noqa
        if legacy_context:
            return get_expected_data_time_for_asset_key(
                legacy_context, will_materialize=not self.true_subset.is_empty
            )
        return None

    def pprint(self, indent=0) -> str:
        ret = f"AutomationResult(label={self.condition.name}, description={self.condition.description}, true={self.true_subset})"
        for child in self.child_results:
            nindent = indent + 4
            ret += f"\n{' '*nindent}{child.pprint(indent=nindent)}"
        return ret

    def __repr__(self) -> str:
        return self.pprint()


def _compute_subset_value_str(subset: SerializableEntitySubset) -> str:
    """Computes a unique string representing a given AssetSubsets. This string will be equal for
    equivalent AssetSubsets.
    """
    if isinstance(subset.value, bool):
        return str(subset.value)
    elif isinstance(subset.value, AllPartitionsSubset):
        return AllPartitionsSubset.__name__
    elif isinstance(subset.value, TimeWindowPartitionsSubset):
        return str(
            [
                (tw.start.timestamp(), tw.end.timestamp())
                for tw in sorted(subset.value.included_time_windows)
            ]
        )
    else:
        return str(list(sorted(subset.subset_value.get_partition_keys())))


def _compute_subset_with_metadata_value_str(subset_with_metadata: AssetSubsetWithMetadata):
    return _compute_subset_value_str(subset_with_metadata.subset) + str(
        sorted(subset_with_metadata.frozen_metadata)
    )



from collections.abc import Sequence
from typing import NamedTuple, Optional

import dagster._check as check
from dagster._serdes import whitelist_for_serdes
from dagster._utils.error import SerializableErrorInfo


class ScheduledExecutionResult:
    pass


@whitelist_for_serdes
class ScheduledExecutionSkipped(
    NamedTuple("_ScheduledExecutionSkipped", []), ScheduledExecutionResult
):
    pass


@whitelist_for_serdes
class ScheduledExecutionFailed(
    NamedTuple(
        "_ScheduledExecutionFailed",
        [("run_id", Optional[str]), ("errors", Sequence[SerializableErrorInfo])],
    ),
    ScheduledExecutionResult,
):
    def __new__(cls, run_id: Optional[str], errors: Sequence[SerializableErrorInfo]):
        return super().__new__(
            cls,
            run_id=check.opt_str_param(run_id, "run_id"),
            errors=check.sequence_param(errors, "errors", of_type=SerializableErrorInfo),
        )


@whitelist_for_serdes
class ScheduledExecutionSuccess(
    NamedTuple("_ScheduledExecutionSuccess", [("run_id", str)]), ScheduledExecutionResult
):
    def __new__(cls, run_id: str):
        return super().__new__(cls, run_id=check.str_param(run_id, "run_id"))



import itertools
from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from typing import AbstractSet, cast  # noqa: UP035

from dagster import _check as check
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._utils.warnings import beta_warning


class ResolvedAssetDependencies:
    """An asset can depend on another asset without specifying the full asset key for the upstream
    asset, if the name and groups match.

    ResolvedAssetDependencies maps these flexible dependencies to precise key-based dependencies.
    """

    def __init__(
        self, assets_defs: Iterable[AssetsDefinition], source_assets: Iterable[SourceAsset]
    ):
        self._deps_by_assets_def_id = resolve_assets_def_deps(assets_defs, source_assets)

    def get_resolved_upstream_asset_keys(
        self, assets_def: AssetsDefinition, asset_key: AssetKey
    ) -> AbstractSet[AssetKey]:
        resolved_keys_by_unresolved_key = self._deps_by_assets_def_id.get(id(assets_def), {})
        return {
            resolved_keys_by_unresolved_key.get(unresolved_dep.asset_key, unresolved_dep.asset_key)
            for unresolved_dep in assets_def.specs_by_key[asset_key].deps
        }

    def get_resolved_asset_key_for_input(
        self, assets_def: AssetsDefinition, input_name: str
    ) -> AssetKey:
        unresolved_asset_key_for_input = assets_def.node_keys_by_input_name[input_name]
        return self._deps_by_assets_def_id.get(id(assets_def), {}).get(
            unresolved_asset_key_for_input, unresolved_asset_key_for_input
        )


def resolve_similar_asset_names(
    target_asset_key: AssetKey,
    asset_keys: Iterable[AssetKey],
) -> Sequence[AssetKey]:
    """Given a target asset key (an upstream dependency which we can't find), produces a list of
    similar asset keys from the list of asset definitions. We use this list to produce a helpful
    error message that can help users debug their asset dependencies.
    """
    similar_names: list[AssetKey] = []

    target_asset_key_split = ("/".join(target_asset_key.path)).split("/")

    for asset_key in asset_keys:
        *target_asset_key_prefix, target_asset_key_name = target_asset_key.path
        *asset_key_prefix, asset_key_name = asset_key.path

        try:
            from rapidfuzz import fuzz

            is_similar_name = bool(
                fuzz.ratio(asset_key_name, target_asset_key_name, score_cutoff=80)
            )
            is_similar_prefix = bool(
                fuzz.ratio(
                    " ".join(asset_key_prefix),
                    " ".join(target_asset_key_prefix),
                    score_cutoff=80,
                )
            )
        except ImportError:
            from difflib import get_close_matches

            is_similar_name = bool(
                get_close_matches(asset_key_name, [target_asset_key_name], cutoff=0.8)
            )
            is_similar_prefix = bool(
                get_close_matches(
                    " ".join(asset_key_prefix), [" ".join(target_asset_key_prefix)], cutoff=0.8
                )
            )

        # Whether the asset key or upstream key has the same prefix and a similar
        # name
        # e.g. [snowflake, elementl, key] and [snowflake, elementl, ey]
        is_same_prefix_similar_name = (
            asset_key_prefix == target_asset_key_prefix and is_similar_name
        )

        # Whether the asset key or upstream key has a similar prefix and the same
        # name
        # e.g. [snowflake, elementl, key] and [nowflake, elementl, key]
        is_similar_prefix_same_name = asset_key_name == target_asset_key_name and is_similar_prefix

        # Whether the asset key or upstream key has one more prefix component than
        # the other, and the same name
        # e.g. [snowflake, elementl, key] and [snowflake, elementl, prod, key]
        is_off_by_one_prefix_component_same_name = (
            asset_key.path[-1] == target_asset_key.path[-1]
            and len(set(asset_key.path).symmetric_difference(set(target_asset_key.path))) == 1
            and max(len(asset_key.path), len(target_asset_key.path)) > 1
        )

        # If the asset key provided has no prefix and the upstream key has
        # the same name but a prefix of any length
        no_prefix_but_is_match_with_prefix = (
            len(target_asset_key.path) == 1 and asset_key.path[-1] == target_asset_key.path[-1]
        )

        matches_slashes_turned_to_prefix_gaps = asset_key.path == target_asset_key_split

        if (
            is_same_prefix_similar_name
            or is_similar_prefix_same_name
            or is_off_by_one_prefix_component_same_name
            or no_prefix_but_is_match_with_prefix
            or matches_slashes_turned_to_prefix_gaps
        ):
            similar_names.append(asset_key)
    return sorted(similar_names, key=lambda key: key.to_string())


def resolve_assets_def_deps(
    assets_defs: Iterable[AssetsDefinition], source_assets: Iterable[SourceAsset]
) -> Mapping[int, Mapping[AssetKey, AssetKey]]:
    """For each AssetsDefinition, resolves its inputs to upstream asset keys. Matches based on either
    of two criteria:
    - The input asset key exactly matches an asset key.
    - The input asset key has one component, that component matches the final component of an asset
        key, and they're both in the same asset group.

    The returned dictionary only contains entries for assets definitions with group-resolved asset
    dependencies.
    """
    group_names_by_key: dict[AssetKey, str] = {}
    for assets_def in assets_defs:
        for spec in assets_def.specs:
            group_names_by_key[spec.key] = check.not_none(spec.group_name)
    for source_asset in source_assets:
        group_names_by_key[source_asset.key] = source_asset.group_name

    all_asset_keys = group_names_by_key.keys()

    asset_keys_by_group_and_name: dict[tuple[str, str], list[AssetKey]] = defaultdict(list)

    for key, group in group_names_by_key.items():
        asset_keys_by_group_and_name[(group, key.path[-1])].append(key)

    warned = False

    result: dict[int, Mapping[AssetKey, AssetKey]] = {}
    for assets_def in assets_defs:
        # If all keys have the same group name, use that
        group_names = {spec.group_name for spec in assets_def.specs}
        group_name = next(iter(group_names)) if len(group_names) == 1 else None

        resolved_keys_by_unresolved_key: dict[AssetKey, AssetKey] = {}
        for input_name, upstream_key in assets_def.keys_by_input_name.items():
            group_and_upstream_name = (group_name, upstream_key.path[-1])
            matching_asset_keys = asset_keys_by_group_and_name.get(
                cast(tuple[str, str], group_and_upstream_name)
            )
            if upstream_key in all_asset_keys:
                pass
            elif (
                group_name is not None
                and len(upstream_key.path) == 1
                and matching_asset_keys
                and len(matching_asset_keys) == 1
                and matching_asset_keys[0] not in assets_def.keys
            ):
                resolved_key = matching_asset_keys[0]
                resolved_keys_by_unresolved_key[upstream_key] = resolved_key

                if not warned:
                    beta_warning(
                        f"Asset {next(iter(assets_def.keys)).to_string()}'s dependency"
                        f" '{upstream_key.path[-1]}' was resolved to upstream asset"
                        f" {resolved_key.to_string()}, because the name matches and they're in the"
                        " same group. This is a beta functionality that may change in a"
                        " future release"
                    )

                    warned = True
            elif not assets_def.node_def.input_def_named(input_name).dagster_type.is_nothing:
                msg = (
                    f"Input asset '{upstream_key.to_string()}' for asset "
                    f"'{next(iter(assets_def.keys)).to_string()}' is not "
                    "produced by any of the provided asset ops and is not one of the provided "
                    "sources."
                )
                similar_names = resolve_similar_asset_names(
                    upstream_key,
                    list(
                        itertools.chain.from_iterable(asset_def.keys for asset_def in assets_defs)
                    ),
                )
                if similar_names:
                    # Arbitrarily limit to 10 similar names to avoid a huge error message
                    subset_similar_names = similar_names[:10]
                    similar_to_string = ", ".join(
                        similar.to_string() for similar in subset_similar_names
                    )
                    msg += f" Did you mean one of the following?\n\t{similar_to_string}"
                raise DagsterInvalidDefinitionError(msg)

        if resolved_keys_by_unresolved_key:
            result[id(assets_def)] = resolved_keys_by_unresolved_key

    return result



from dagster._core.executor.step_delegating.step_delegating_executor import (
    StepDelegatingExecutor as StepDelegatingExecutor,
)
from dagster._core.executor.step_delegating.step_handler import (
    CheckStepHealthResult as CheckStepHealthResult,
    StepHandler as StepHandler,
    StepHandlerContext as StepHandlerContext,
)



import enum
from collections.abc import Iterable
from typing import NamedTuple, Optional, cast

import dagster._check as check
from dagster._core.definitions.asset_check_evaluation import AssetCheckEvaluation
from dagster._core.definitions.asset_key import AssetCheckKey
from dagster._core.events.log import DagsterEventType, EventLogEntry
from dagster._core.loader import LoadableBy, LoadingContext
from dagster._core.storage.dagster_run import DagsterRunStatus, RunRecord
from dagster._serdes.serdes import deserialize_value
from dagster._time import utc_datetime_from_naive


class AssetCheckInstanceSupport(enum.Enum):
    """Reasons why a dagster instance might not support checks."""

    SUPPORTED = "SUPPORTED"
    NEEDS_MIGRATION = "NEEDS_MIGRATION"
    NEEDS_AGENT_UPGRADE = "NEEDS_AGENT_UPGRADE"


# We store a limit set of statuses in the database, and then resolve them to the actual status
# at read time. This is because the write path is to store a planned event (which creates a row
# with PLANNED status) then update the row when we get the check result. But if the check never
# runs, the row stays in PLANNED status.
class AssetCheckExecutionRecordStatus(enum.Enum):
    PLANNED = "PLANNED"
    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"  # explicit fail result


COMPLETED_ASSET_CHECK_EXECUTION_RECORD_STATUSES = {
    AssetCheckExecutionRecordStatus.SUCCEEDED,
    AssetCheckExecutionRecordStatus.FAILED,
}


class AssetCheckExecutionResolvedStatus(enum.Enum):
    IN_PROGRESS = "IN_PROGRESS"
    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"  # explicit fail result
    EXECUTION_FAILED = "EXECUTION_FAILED"  # hit some exception
    SKIPPED = "SKIPPED"  # the run finished, didn't fail, but the check didn't execute


class AssetCheckExecutionRecord(
    NamedTuple(
        "_AssetCheckExecutionRecord",
        [
            ("key", AssetCheckKey),
            ("id", int),
            ("run_id", str),
            ("status", AssetCheckExecutionRecordStatus),
            # Either an AssetCheckEvaluationPlanned or AssetCheckEvaluation event.
            # Optional for backwards compatibility, before we started storing planned events.
            # Old records won't have an event if the status is PLANNED.
            ("event", Optional[EventLogEntry]),
            ("create_timestamp", float),
        ],
    ),
    LoadableBy[AssetCheckKey],
):
    def __new__(
        cls,
        key: AssetCheckKey,
        id: int,
        run_id: str,
        status: AssetCheckExecutionRecordStatus,
        event: Optional[EventLogEntry],
        create_timestamp: float,
    ):
        check.inst_param(key, "key", AssetCheckKey)
        check.int_param(id, "id")
        check.str_param(run_id, "run_id")
        check.inst_param(status, "status", AssetCheckExecutionRecordStatus)
        check.opt_inst_param(event, "event", EventLogEntry)
        check.float_param(create_timestamp, "create_timestamp")

        event_type = event.dagster_event_type if event else None
        if status == AssetCheckExecutionRecordStatus.PLANNED:
            check.invariant(
                event is None or event_type == DagsterEventType.ASSET_CHECK_EVALUATION_PLANNED,
                f"The asset check row status is PLANNED, but the event is type {event_type} instead"
                " of ASSET_CHECK_EVALUATION_PLANNED",
            )
        elif status in [
            AssetCheckExecutionRecordStatus.FAILED,
            AssetCheckExecutionRecordStatus.SUCCEEDED,
        ]:
            check.invariant(
                event_type == DagsterEventType.ASSET_CHECK_EVALUATION,
                f"The asset check row status is {status}, but the event is type"
                f" {event_type} instead of ASSET_CHECK_EVALUATION",
            )

        return super().__new__(
            cls,
            key=key,
            id=id,
            run_id=run_id,
            status=status,
            event=event,
            create_timestamp=create_timestamp,
        )

    @property
    def evaluation(self) -> Optional[AssetCheckEvaluation]:
        if self.event and self.event.dagster_event:
            return cast(
                AssetCheckEvaluation,
                self.event.dagster_event.event_specific_data,
            )
        return None

    @classmethod
    def from_db_row(cls, row, key: AssetCheckKey) -> "AssetCheckExecutionRecord":
        return cls(
            key=key,
            id=row["id"],
            run_id=row["run_id"],
            status=AssetCheckExecutionRecordStatus(row["execution_status"]),
            event=(
                deserialize_value(row["evaluation_event"], EventLogEntry)
                if row["evaluation_event"]
                else None
            ),
            create_timestamp=utc_datetime_from_naive(row["create_timestamp"]).timestamp(),
        )

    @classmethod
    def _blocking_batch_load(
        cls, keys: Iterable[AssetCheckKey], context: LoadingContext
    ) -> Iterable[Optional["AssetCheckExecutionRecord"]]:
        records_by_key = context.instance.event_log_storage.get_latest_asset_check_execution_by_key(
            list(keys)
        )
        return [records_by_key.get(key) for key in keys]

    async def resolve_status(
        self, loading_context: LoadingContext
    ) -> AssetCheckExecutionResolvedStatus:
        if self.status == AssetCheckExecutionRecordStatus.SUCCEEDED:
            return AssetCheckExecutionResolvedStatus.SUCCEEDED
        elif self.status == AssetCheckExecutionRecordStatus.FAILED:
            return AssetCheckExecutionResolvedStatus.FAILED
        elif self.status == AssetCheckExecutionRecordStatus.PLANNED:
            # Asset checks stay in PLANNED status until the evaluation event arrives.
            # Check if the run is still active, and if not, return the actual status.
            run_record = await RunRecord.gen(loading_context, self.run_id)
            if not run_record:
                # Run deleted
                return AssetCheckExecutionResolvedStatus.SKIPPED

            run = run_record.dagster_run
            if run.is_finished:
                return (
                    AssetCheckExecutionResolvedStatus.EXECUTION_FAILED
                    if run.status == DagsterRunStatus.FAILURE
                    else AssetCheckExecutionResolvedStatus.SKIPPED
                )
            else:
                return AssetCheckExecutionResolvedStatus.IN_PROGRESS
        else:
            check.failed(f"Unexpected status {self.status}")

    async def targets_latest_materialization(self, loading_context: LoadingContext) -> bool:
        from dagster._core.storage.event_log.base import AssetRecord

        resolved_status = await self.resolve_status(loading_context)
        if resolved_status == AssetCheckExecutionResolvedStatus.IN_PROGRESS:
            # all in-progress checks execute against the latest version
            return True

        asset_record = await AssetRecord.gen(loading_context, self.key.asset_key)
        latest_materialization = (
            asset_record.asset_entry.last_materialization_record if asset_record else None
        )
        if not latest_materialization:
            # no previous materialization, so it's executing against the lastest version
            return True

        latest_materialization_run_id = latest_materialization.event_log_entry.run_id
        if latest_materialization_run_id == self.run_id:
            # part of the same run
            return True

        if resolved_status in [
            AssetCheckExecutionResolvedStatus.SUCCEEDED,
            AssetCheckExecutionResolvedStatus.FAILED,
        ]:
            evaluation = check.not_none(self.evaluation)
            if not evaluation.target_materialization_data:
                # check ran before the materialization was created
                return False
            else:
                # check matches the latest materialization id
                return (
                    evaluation.target_materialization_data.storage_id
                    == latest_materialization.storage_id
                )
        elif resolved_status in [
            AssetCheckExecutionResolvedStatus.EXECUTION_FAILED,
            AssetCheckExecutionResolvedStatus.SKIPPED,
        ]:
            # the evaluation didn't complete, so we don't have target_materialization_data, so check if
            # the check's run executed after the materializations as a fallback
            latest_materialization_run_record = await RunRecord.gen(
                loading_context, latest_materialization_run_id
            )
            check_run_record = await RunRecord.gen(loading_context, self.run_id)
            return bool(
                latest_materialization_run_record
                and check_run_record
                and check_run_record.create_timestamp
                > latest_materialization_run_record.create_timestamp
            )
        else:
            check.failed(f"Unexpected check status {resolved_status}")



import json
from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from inspect import isfunction
from typing import TYPE_CHECKING, Any, Callable, Optional, Union, cast

import dagster._check as check
from dagster._config.pythonic_config import (
    ConfigurableIOManagerFactoryResourceDefinition,
    ConfigurableResourceFactoryResourceDefinition,
)
from dagster._core.definitions.asset_graph import AssetGraph
from dagster._core.definitions.asset_job import (
    IMPLICIT_ASSET_JOB_NAME,
    get_base_asset_job_lambda,
    is_reserved_asset_job_name,
)
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.automation_condition_sensor_definition import (
    AutomationConditionSensorDefinition,
)
from dagster._core.definitions.base_asset_graph import BaseAssetGraph
from dagster._core.definitions.executor_definition import ExecutorDefinition
from dagster._core.definitions.graph_definition import GraphDefinition
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.logger_definition import LoggerDefinition
from dagster._core.definitions.multi_dimensional_partitions import MultiPartitionsDefinition
from dagster._core.definitions.partition import (
    DynamicPartitionsDefinition,
    PartitionsDefinition,
    StaticPartitionsDefinition,
)
from dagster._core.definitions.partitioned_schedule import (
    UnresolvedPartitionedAssetScheduleDefinition,
)
from dagster._core.definitions.repository_definition.repository_data import CachingRepositoryData
from dagster._core.definitions.repository_definition.valid_definitions import (
    VALID_REPOSITORY_DATA_DICT_KEYS,
    RepositoryElementDefinition,
)
from dagster._core.definitions.resource_definition import ResourceDefinition
from dagster._core.definitions.schedule_definition import ScheduleDefinition
from dagster._core.definitions.sensor_definition import SensorDefinition
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.definitions.time_window_partitions import TimeWindowPartitionsDefinition
from dagster._core.definitions.unresolved_asset_job_definition import UnresolvedAssetJobDefinition
from dagster._core.definitions.utils import get_default_automation_condition_sensor
from dagster._core.errors import DagsterInvalidDefinitionError

if TYPE_CHECKING:
    from dagster._core.definitions.asset_check_spec import AssetCheckKey
    from dagster._core.definitions.events import AssetKey

# We throw an error if the user attaches an instance of a custom `PartitionsDefinition` subclass to
# a definition-- we can't support custom PartitionsDefinition subclasses due to us needing to load
# them in the host process.
VALID_PARTITIONS_DEFINITION_CLASSES = (
    StaticPartitionsDefinition,
    DynamicPartitionsDefinition,
    TimeWindowPartitionsDefinition,
    MultiPartitionsDefinition,
)


def _find_env_vars(config_entry: Any) -> set[str]:
    """Given a part of a config dictionary, return a set of environment variables that are used in
    that part of the config.
    """
    # Actual env var entry
    if isinstance(config_entry, Mapping) and set(config_entry.keys()) == {"env"}:
        return {config_entry["env"]}
    # Recurse into dictionary of config items
    elif isinstance(config_entry, Mapping):
        return set().union(*[_find_env_vars(v) for v in config_entry.values()])
    # Recurse into list of config items
    elif isinstance(config_entry, list):
        return set().union(*[_find_env_vars(v) for v in config_entry])

    # Otherwise, raw config value which is not an env var, so return empty set
    return set()


def _env_vars_from_resource_defaults(resource_def: ResourceDefinition) -> set[str]:
    """Given a resource definition, return a set of environment variables that are used in the
    resource's default config. This is used to extract environment variables from the top-level
    resources in a Definitions object.
    """
    from dagster._core.execution.build_resources import wrap_resource_for_execution

    config_schema_default = cast(
        Mapping[str, Any],
        (
            json.loads(resource_def.config_schema.default_value_as_json_str)
            if resource_def.config_schema.default_provided
            else {}
        ),
    )

    env_vars = _find_env_vars(config_schema_default)

    if isinstance(
        resource_def,
        (
            ConfigurableIOManagerFactoryResourceDefinition,
            ConfigurableResourceFactoryResourceDefinition,
        ),
    ):
        nested_resources = resource_def.nested_resources
        for nested_resource in nested_resources.values():
            env_vars = env_vars.union(
                _env_vars_from_resource_defaults(wrap_resource_for_execution(nested_resource))
            )

    return env_vars


def _resolve_unresolved_job_def_lambda(
    unresolved_job_def: UnresolvedAssetJobDefinition,
    asset_graph: AssetGraph,
    default_executor_def: Optional[ExecutorDefinition],
    top_level_resources: Optional[Mapping[str, ResourceDefinition]],
    default_logger_defs: Optional[Mapping[str, LoggerDefinition]],
) -> Callable[[], JobDefinition]:
    def resolve_unresolved_job_def() -> JobDefinition:
        job_def = unresolved_job_def.resolve(
            asset_graph=asset_graph,
            default_executor_def=default_executor_def,
            resource_defs=top_level_resources,
        )
        return _process_resolved_job(job_def, default_executor_def, default_logger_defs)

    return resolve_unresolved_job_def


def _process_resolved_job(
    job_def: JobDefinition,
    default_executor_def: Optional[ExecutorDefinition],
    default_logger_defs: Optional[Mapping[str, LoggerDefinition]],
) -> JobDefinition:
    job_def.validate_resource_requirements_satisfied()

    if default_executor_def and not job_def.has_specified_executor:
        job_def = job_def.with_executor_def(default_executor_def)

    if default_logger_defs and not job_def.has_specified_loggers:
        job_def = job_def.with_logger_defs(default_logger_defs)

    return job_def


def build_caching_repository_data_from_list(
    repository_definitions: Sequence[RepositoryElementDefinition],
    default_executor_def: Optional[ExecutorDefinition] = None,
    default_logger_defs: Optional[Mapping[str, LoggerDefinition]] = None,
    top_level_resources: Optional[Mapping[str, ResourceDefinition]] = None,
    resource_key_mapping: Optional[Mapping[int, str]] = None,
) -> CachingRepositoryData:
    from dagster._core.definitions import AssetsDefinition
    from dagster._core.definitions.partitioned_schedule import (
        UnresolvedPartitionedAssetScheduleDefinition,
    )

    schedule_and_sensor_names: set[str] = set()
    jobs: dict[str, Union[JobDefinition, Callable[[], JobDefinition]]] = {}
    coerced_graphs: dict[str, JobDefinition] = {}
    unresolved_jobs: dict[str, UnresolvedAssetJobDefinition] = {}
    schedules: dict[str, ScheduleDefinition] = {}
    unresolved_partitioned_asset_schedules: dict[
        str, UnresolvedPartitionedAssetScheduleDefinition
    ] = {}
    sensors: dict[str, SensorDefinition] = {}
    assets_defs: list[AssetsDefinition] = []
    asset_keys: set[AssetKey] = set()
    asset_check_keys: set[AssetCheckKey] = set()
    source_assets: list[SourceAsset] = []
    asset_checks_defs: list[AssetsDefinition] = []
    partitions_defs: set[PartitionsDefinition] = set()
    for definition in repository_definitions:
        if isinstance(definition, JobDefinition):
            if (
                definition.name in jobs and jobs[definition.name] != definition
            ) or definition.name in unresolved_jobs:
                raise DagsterInvalidDefinitionError(
                    f"Duplicate job definition found for {definition.describe_target()}"
                )
            if is_reserved_asset_job_name(definition.name):
                raise DagsterInvalidDefinitionError(
                    f"Attempted to provide job called {definition.name} to repository, which "
                    "is a reserved name. Please rename the job."
                )
            if definition.partitions_def is not None:
                partitions_defs.add(definition.partitions_def)
            jobs[definition.name] = definition
        elif isinstance(definition, SensorDefinition):
            if definition.name in schedule_and_sensor_names:
                raise DagsterInvalidDefinitionError(
                    f"Duplicate definition found for {definition.name}"
                )
            schedule_and_sensor_names.add(definition.name)
            sensors[definition.name] = definition
        elif isinstance(definition, ScheduleDefinition):
            if definition.name in schedule_and_sensor_names:
                raise DagsterInvalidDefinitionError(
                    f"Duplicate definition found for {definition.name}"
                )
            schedule_and_sensor_names.add(definition.name)

            schedules[definition.name] = definition

        elif isinstance(definition, UnresolvedPartitionedAssetScheduleDefinition):
            if definition.name in schedule_and_sensor_names:
                raise DagsterInvalidDefinitionError(
                    f"Duplicate definition found for {definition.name}"
                )
            schedule_and_sensor_names.add(definition.name)

            unresolved_partitioned_asset_schedules[definition.name] = definition
        elif isinstance(definition, GraphDefinition):
            coerced = definition.coerce_to_job()
            if coerced.name in jobs:
                raise DagsterInvalidDefinitionError(
                    f"Duplicate job definition found for graph '{coerced.name}'"
                )
            jobs[coerced.name] = coerced
            coerced_graphs[coerced.name] = coerced
        elif isinstance(definition, UnresolvedAssetJobDefinition):
            if definition.name in jobs or definition.name in unresolved_jobs:
                raise DagsterInvalidDefinitionError(
                    f"Duplicate definition found for unresolved job '{definition.name}'"
                )
            # we can only resolve these once we have all assets
            if definition.partitions_def is not None:
                partitions_defs.add(definition.partitions_def)
            unresolved_jobs[definition.name] = definition
        elif (
            isinstance(definition, AssetsDefinition)
            and definition.has_check_keys
            and not definition.has_keys
        ):
            for key in definition.check_keys:
                if key in asset_check_keys:
                    raise DagsterInvalidDefinitionError(f"Duplicate asset check key: {key}")
            asset_check_keys.update(definition.check_keys)
            asset_checks_defs.append(definition)
        elif isinstance(definition, AssetsDefinition):
            for spec in definition.specs:
                if spec.key in asset_keys:
                    raise DagsterInvalidDefinitionError(f"Duplicate asset key: {spec.key}")

                if spec.partitions_def is not None:
                    partitions_defs.add(spec.partitions_def)
            for key in definition.check_keys:
                if key in asset_check_keys:
                    raise DagsterInvalidDefinitionError(f"Duplicate asset check key: {key}")

            asset_keys.update(definition.keys)
            asset_check_keys.update(definition.check_keys)
            assets_defs.append(definition)
        elif isinstance(definition, SourceAsset):
            source_assets.append(definition)
            asset_keys.add(definition.key)
        else:
            check.failed(f"Unexpected repository entry {definition}")

    for partitions_def in partitions_defs:
        _validate_partitions_definition(partitions_def)

    source_assets_by_key = {source_asset.key: source_asset for source_asset in source_assets}
    assets_defs_by_key = {key: asset for asset in assets_defs for key in asset.keys}
    asset_checks_defs_by_key = {
        key: checks_def for checks_def in asset_checks_defs for key in checks_def.check_keys
    }

    for name, sensor_def in sensors.items():
        for target in sensor_def.targets:
            if target.has_job_def:
                _process_and_validate_target_job(sensor_def, unresolved_jobs, jobs, target.job_def)
            if target.assets_defs:
                _process_and_validate_target_assets(
                    sensor_def, assets_defs_by_key, source_assets_by_key, target.assets_defs
                )

    for name, schedule_def in schedules.items():
        if schedule_def.target.has_job_def:
            _process_and_validate_target_job(
                schedule_def, unresolved_jobs, jobs, schedule_def.target.job_def
            )
        if schedule_def.target.assets_defs:
            _process_and_validate_target_assets(
                schedule_def,
                assets_defs_by_key,
                source_assets_by_key,
                schedule_def.target.assets_defs,
            )

    if unresolved_partitioned_asset_schedules:
        for (
            name,
            unresolved_partitioned_asset_schedule,
        ) in unresolved_partitioned_asset_schedules.items():
            _process_and_validate_target_job(
                unresolved_partitioned_asset_schedule,
                unresolved_jobs,
                jobs,
                unresolved_partitioned_asset_schedule.job,
            )

    assets_without_keys = [ad for ad in assets_defs if not ad.keys]
    asset_graph = AssetGraph.from_assets(
        [
            # Ensure that the same AssetsDefinition doesn't need to be considered twice by the graph if
            # it produces multiple asset keys
            *set(assets_defs_by_key.values()),
            *assets_without_keys,
            *asset_checks_defs,
            *source_assets_by_key.values(),  # only ever one key per source asset so no need to dedupe
        ]
    )
    # add a default automation condition sensor if necessary
    default_automation_condition_sensor = get_default_automation_condition_sensor(
        list(sensors.values()), asset_graph
    )
    if default_automation_condition_sensor:
        sensors[default_automation_condition_sensor.name] = default_automation_condition_sensor

    if assets_defs or asset_checks_defs or source_assets:
        jobs[IMPLICIT_ASSET_JOB_NAME] = get_base_asset_job_lambda(
            asset_graph=asset_graph,
            executor_def=default_executor_def,
            resource_defs=top_level_resources,
            logger_defs=default_logger_defs,
        )

    _validate_auto_materialize_sensors(sensors.values(), asset_graph)

    # resolve all the UnresolvedAssetJobDefinitions using the full set of assets
    # resolving jobs is potentially time-consuming if there are many of them,
    # so do the resolving lazily in a lambda
    if unresolved_jobs:
        for name, unresolved_job_def in unresolved_jobs.items():
            jobs[name] = _resolve_unresolved_job_def_lambda(
                unresolved_job_def,
                asset_graph,
                default_executor_def,
                top_level_resources,
                default_logger_defs,
            )

    jobs = {
        name: (
            job_def
            if isfunction(job_def)
            else _process_resolved_job(
                cast(JobDefinition, job_def), default_executor_def, default_logger_defs
            )
        )
        for name, job_def in jobs.items()
    }

    top_level_resources = top_level_resources or {}

    utilized_env_vars: dict[str, set[str]] = defaultdict(set)

    for resource_key, resource_def in top_level_resources.items():
        used_env_vars = _env_vars_from_resource_defaults(resource_def)
        for env_var in used_env_vars:
            utilized_env_vars[env_var].add(resource_key)

    return CachingRepositoryData(
        jobs=jobs,
        schedules=schedules,
        sensors=sensors,
        source_assets_by_key=source_assets_by_key,
        assets_defs_by_key=assets_defs_by_key,
        asset_checks_defs_by_key=asset_checks_defs_by_key,
        top_level_resources=top_level_resources or {},
        utilized_env_vars=utilized_env_vars,
        unresolved_partitioned_asset_schedules=unresolved_partitioned_asset_schedules,
    )


def build_caching_repository_data_from_dict(
    repository_definitions: dict[str, dict[str, Any]],
) -> "CachingRepositoryData":
    check.dict_param(repository_definitions, "repository_definitions", key_type=str)
    check.invariant(
        set(repository_definitions.keys()).issubset(VALID_REPOSITORY_DATA_DICT_KEYS),
        "Bad dict: must not contain keys other than {{{valid_keys}}}: found {bad_keys}.".format(
            valid_keys=", ".join([f"'{key}'" for key in VALID_REPOSITORY_DATA_DICT_KEYS]),
            bad_keys=", ".join(
                [
                    "'{key}'"
                    for key in repository_definitions.keys()
                    if key not in VALID_REPOSITORY_DATA_DICT_KEYS
                ]
            ),
        ),
    )

    for key in VALID_REPOSITORY_DATA_DICT_KEYS:
        if key not in repository_definitions:
            repository_definitions[key] = {}

    duplicate_keys = set(repository_definitions["schedules"].keys()).intersection(
        set(repository_definitions["sensors"].keys())
    )
    if duplicate_keys:
        raise DagsterInvalidDefinitionError(
            "Duplicate definitions between schedules and sensors found for keys:"
            f" {', '.join(duplicate_keys)}"
        )

    for key, raw_job_def in repository_definitions["jobs"].items():
        if isinstance(raw_job_def, GraphDefinition):
            repository_definitions["jobs"][key] = raw_job_def.coerce_to_job()
        elif isinstance(raw_job_def, UnresolvedAssetJobDefinition):
            repository_definitions["jobs"][key] = raw_job_def.resolve(
                # TODO: https://github.com/dagster-io/dagster/issues/8263
                asset_graph=AssetGraph.from_assets([]),
                default_executor_def=None,
            )
        elif not isinstance(raw_job_def, JobDefinition) and not isfunction(raw_job_def):
            raise DagsterInvalidDefinitionError(
                f"Object mapped to {key} is not an instance of JobDefinition or GraphDefinition."
            )

    # Late validate all jobs' resource requirements are satisfied, since
    # they may not be applied until now
    for job_def in repository_definitions["jobs"].values():
        if isinstance(job_def, JobDefinition):
            job_def.validate_resource_requirements_satisfied()

    return CachingRepositoryData(
        **repository_definitions,
        source_assets_by_key={},
        assets_defs_by_key={},
        asset_checks_defs_by_key={},
        top_level_resources={},
        utilized_env_vars={},
        unresolved_partitioned_asset_schedules={},
    )


def _process_and_validate_target_job(
    instigator_def: Union[
        SensorDefinition, ScheduleDefinition, UnresolvedPartitionedAssetScheduleDefinition
    ],
    unresolved_jobs: dict[str, UnresolvedAssetJobDefinition],
    jobs: dict[str, Union[JobDefinition, Callable[[], JobDefinition]]],
    job_def: Union[JobDefinition, UnresolvedAssetJobDefinition],
):
    """This function modifies the state of unresolved_jobs, and jobs."""
    targeter = (
        f"schedule '{instigator_def.name}'"
        if isinstance(instigator_def, ScheduleDefinition)
        else f"sensor '{instigator_def.name}'"
    )
    if isinstance(job_def, UnresolvedAssetJobDefinition):
        if job_def.name not in unresolved_jobs:
            # Since this is an unresolved job we have to resolve, it is not possible to
            # be the same definition by reference equality
            if job_def.name in jobs:
                raise DagsterInvalidDefinitionError(
                    _get_error_msg_for_target_conflict(
                        targeter, "unresolved asset job", job_def.name, "job"
                    )
                )
        elif unresolved_jobs[job_def.name].selection != job_def.selection:
            raise DagsterInvalidDefinitionError(
                _get_error_msg_for_target_conflict(
                    targeter, "unresolved asset job", job_def.name, "unresolved asset job"
                )
            )
        unresolved_jobs[job_def.name] = job_def
    else:
        if job_def.name in jobs and jobs[job_def.name].graph != job_def.graph:
            dupe_target_type = "unresolved asset job" if job_def.name in unresolved_jobs else "job"
            raise DagsterInvalidDefinitionError(
                _get_error_msg_for_target_conflict(targeter, "job", job_def.name, dupe_target_type)
            )
        jobs[job_def.name] = job_def


def _process_and_validate_target_assets(
    instigator_def: Union[
        SensorDefinition, ScheduleDefinition, UnresolvedPartitionedAssetScheduleDefinition
    ],
    assets_defs_by_key: dict["AssetKey", AssetsDefinition],
    source_assets_by_key: dict["AssetKey", SourceAsset],
    target_assets_defs: Sequence[Union[AssetsDefinition, SourceAsset]],
) -> None:
    for ad in target_assets_defs:
        keys = ad.keys if isinstance(ad, AssetsDefinition) else [ad.key]
        for key in keys:
            if isinstance(ad, AssetsDefinition):
                existing_ad = assets_defs_by_key.get(key)
                if not existing_ad:
                    assets_defs_by_key[key] = ad
                elif not existing_ad.computation == ad.computation:
                    raise DagsterInvalidDefinitionError(
                        _get_error_msg_for_target_asset_conflict(
                            _get_instigator_str(instigator_def), key.to_user_string()
                        )
                    )
            elif isinstance(ad, SourceAsset):
                existing_sa = source_assets_by_key.get(key)
                if not existing_sa:
                    source_assets_by_key[key] = ad
                elif not ad == existing_sa:
                    raise DagsterInvalidDefinitionError(
                        _get_error_msg_for_target_asset_conflict(
                            _get_instigator_str(instigator_def), key.to_user_string()
                        )
                    )
            else:
                check.failed(f"Unexpected asset definition {ad}")


def _validate_auto_materialize_sensors(
    sensors: Iterable[SensorDefinition], asset_graph: BaseAssetGraph
) -> None:
    """Raises an error if two or more automation policy sensors target the same asset."""
    sensor_names_by_asset_key: dict[AssetKey, str] = {}
    for sensor in sensors:
        if isinstance(sensor, AutomationConditionSensorDefinition):
            asset_keys = sensor.asset_selection.resolve(asset_graph)
            for asset_key in asset_keys:
                if asset_key in sensor_names_by_asset_key:
                    raise DagsterInvalidDefinitionError(
                        f"Automation policy sensors '{sensor.name}' and "
                        f"'{sensor_names_by_asset_key[asset_key]}' have overlapping asset "
                        f"selections: they both target '{asset_key.to_user_string()}'. Each asset "
                        "must only be targeted by one automation policy sensor."
                    )
                else:
                    sensor_names_by_asset_key[asset_key] = sensor.name


def _validate_partitions_definition(partitions_def: PartitionsDefinition) -> None:
    if not isinstance(partitions_def, VALID_PARTITIONS_DEFINITION_CLASSES):
        valid_names = ", ".join([cls.__name__ for cls in VALID_PARTITIONS_DEFINITION_CLASSES])
        raise DagsterInvalidDefinitionError(
            "Custom PartitionsDefinition subclasses are not supported. All passed-in PartitionsDefinition"
            f" objects must be an instance of one of ({valid_names})."
            f" Found instance of {type(partitions_def).__name__}",
        )


def _get_instigator_str(
    instigator_def: Union[
        SensorDefinition, ScheduleDefinition, UnresolvedPartitionedAssetScheduleDefinition
    ],
) -> str:
    return (
        f"schedule '{instigator_def.name}'"
        if isinstance(
            instigator_def,
            (ScheduleDefinition, UnresolvedPartitionedAssetScheduleDefinition),
        )
        else f"sensor '{instigator_def.name}'"
    )


def _get_error_msg_for_target_conflict(
    instigator: str, target_type: str, target_name: str, dupe_target_type: str
) -> str:
    return (
        f"{instigator} targets {target_type} '{target_name}', but a different {dupe_target_type} with"
        " the same name was provided. Disambiguate between these by providing a separate name to"
        " one of them."
    )


def _get_error_msg_for_target_asset_conflict(instigator: str, target_name: str) -> str:
    return (
        f"{instigator} targets asset '{target_name}', but a different asset with"
        " the same key was provided. Assets must have unique keys."
    )



import time
from collections.abc import Iterator, Mapping, Sequence
from types import TracebackType
from typing import Any, Callable, Optional, Union, cast

from typing_extensions import Self

import dagster._check as check
from dagster._core.errors import (
    DagsterExecutionInterruptedError,
    DagsterInvariantViolationError,
    DagsterUnknownStepStateError,
)
from dagster._core.events import DagsterEvent
from dagster._core.execution.context.system import (
    IPlanContext,
    PlanExecutionContext,
    PlanOrchestrationContext,
)
from dagster._core.execution.plan.instance_concurrency_context import InstanceConcurrencyContext
from dagster._core.execution.plan.outputs import StepOutputData, StepOutputHandle
from dagster._core.execution.plan.plan import ExecutionPlan
from dagster._core.execution.plan.state import KnownExecutionState
from dagster._core.execution.plan.step import ExecutionStep
from dagster._core.execution.retries import RetryMode, RetryState
from dagster._core.storage.tags import GLOBAL_CONCURRENCY_TAG, PRIORITY_TAG
from dagster._utils.interrupts import pop_captured_interrupt
from dagster._utils.tags import TagConcurrencyLimitsCounter


def _default_sort_key(step: ExecutionStep) -> float:
    return int(step.tags.get(PRIORITY_TAG, 0)) * -1


def _pool_key_for_step(step: ExecutionStep) -> Optional[str]:
    # for backwards compatibility, we also check the tags
    return step.pool or step.tags.get(GLOBAL_CONCURRENCY_TAG)


CONCURRENCY_CLAIM_BLOCKED_INTERVAL = 1
CONCURRENCY_CLAIM_MESSAGE_INTERVAL = 300


class ActiveExecution:
    """State machine used to track progress through execution of an ExecutionPlan."""

    def __init__(
        self,
        execution_plan: ExecutionPlan,
        retry_mode: RetryMode,
        sort_key_fn: Optional[Callable[[ExecutionStep], float]] = None,
        max_concurrent: Optional[int] = None,
        tag_concurrency_limits: Optional[list[dict[str, Any]]] = None,
        instance_concurrency_context: Optional[InstanceConcurrencyContext] = None,
    ):
        self._plan: ExecutionPlan = check.inst_param(
            execution_plan, "execution_plan", ExecutionPlan
        )
        self._retry_mode = check.inst_param(retry_mode, "retry_mode", RetryMode)
        self._retry_state = self._plan.known_state.get_retry_state()
        self._instance_concurrency_context = instance_concurrency_context

        self._sort_key_fn: Callable[[ExecutionStep], float] = (
            check.opt_callable_param(
                sort_key_fn,
                "sort_key_fn",
            )
            or _default_sort_key
        )

        self._max_concurrent = check.opt_int_param(max_concurrent, "max_concurrent")
        self._tag_concurrency_limits = check.opt_list_param(
            tag_concurrency_limits, "tag_concurrency_limits"
        )

        self._context_guard: bool = False  # Prevent accidental direct use

        # We decide what steps to skip based on what outputs are yielded by upstream steps
        self._step_outputs: set[StepOutputHandle] = set(self._plan.known_state.ready_outputs)

        # All steps to be executed start out here in _pending
        self._pending: dict[str, set[str]] = dict(self._plan.get_executable_step_deps())

        # track mapping keys from DynamicOutputs, step_key, output_name -> list of keys
        # to _gathering while in flight
        self._gathering_dynamic_outputs: dict[str, Mapping[str, Optional[list[str]]]] = {}
        # then on resolution move to _completed
        self._completed_dynamic_outputs: dict[str, Mapping[str, Optional[Sequence[str]]]] = (
            dict(self._plan.known_state.dynamic_mappings) if self._plan.known_state else {}
        )
        self._new_dynamic_mappings: bool = False

        # track which upstream deps caused a step to skip
        self._skipped_deps: dict[str, Sequence[str]] = {}

        # steps move in to these buckets as a result of _update calls
        self._executable: list[str] = []
        self._pending_skip: list[str] = []
        self._pending_retry: list[str] = []
        self._pending_abandon: list[str] = []
        self._waiting_to_retry: dict[str, float] = {}
        self._messaged_concurrency_slots: dict[str, float] = {}

        # then are considered _in_flight when vended via get_steps_to_*
        self._in_flight: set[str] = set()

        # and finally their terminal state is tracked by these sets, via mark_*
        self._success: set[str] = set()
        self._failed: set[str] = set()
        self._skipped: set[str] = set()
        self._abandoned: set[str] = set()

        # see verify_complete
        self._unknown_state: set[str] = set()

        self._interrupted: bool = False

        # Start the show by loading _executable with the set of _pending steps that have no deps
        self._update()

    def __enter__(self) -> Self:
        self._context_guard = True
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self._context_guard = False

        # Exiting due to exception, return to allow exception to bubble
        if exc_type or exc_value or traceback:
            return

        if not self.is_complete:
            # generate the state string before exiting the concurrency context
            state_str = self._pending_state_str()
        else:
            state_str = ""

        if not self.is_complete:
            if self._interrupted:
                raise DagsterExecutionInterruptedError(
                    f"Execution was interrupted before completing the execution plan. {state_str}"
                )
            else:
                raise DagsterInvariantViolationError(
                    f"Execution finished without completing the execution plan. {state_str}"
                )

        # See verify_complete - steps for which we did not observe a failure/success event are in an unknown
        # state so we raise to ensure pipeline failure.
        if len(self._unknown_state) > 0:
            if self._interrupted:
                raise DagsterExecutionInterruptedError(
                    f"Execution exited with steps {self._unknown_state} in an unknown state after "
                    "being interrupted."
                )
            else:
                raise DagsterUnknownStepStateError(
                    f"Execution exited with steps {self._unknown_state} in an unknown state to this"
                    " process.\nThis was likely caused by losing communication with the process"
                    " performing step execution."
                )

    def _pending_state_str(self) -> str:
        assert not self.is_complete
        pending_action = (
            self._executable + self._pending_abandon + self._pending_retry + self._pending_skip
        )
        return "{pending_str}{in_flight_str}{action_str}{retry_str}{claim_str}".format(
            in_flight_str=f"\nSteps still in flight: {self._in_flight}" if self._in_flight else "",
            pending_str=(
                f"\nSteps pending processing: {self._pending.keys()}" if self._pending else ""
            ),
            action_str=f"\nSteps pending action: {pending_action}" if pending_action else "",
            retry_str=(
                f"\nSteps waiting to retry: {self._waiting_to_retry.keys()}"
                if self._waiting_to_retry
                else ""
            ),
            claim_str=(
                "\nSteps waiting to claim:"
                f" {self._instance_concurrency_context.pending_claim_steps()}"
                if self._instance_concurrency_context
                and self._instance_concurrency_context.has_pending_claims()
                else ""
            ),
        )

    def _should_skip_step(self, step_key: str, successful_or_skipped_steps: set[str]) -> bool:
        step = self.get_step_by_key(step_key)
        for step_input in step.step_inputs:
            missing_source_handles = []

            for source_handle in step_input.get_step_output_handle_dependencies():
                if (
                    source_handle.step_key in successful_or_skipped_steps
                    and source_handle not in self._step_outputs
                ):
                    missing_source_handles.append(source_handle)

            if missing_source_handles:
                if len(missing_source_handles) == len(
                    step_input.get_step_output_handle_dependencies()
                ):
                    self._skipped_deps[step_key] = [
                        f"{h.step_key}.{h.output_name}" for h in missing_source_handles
                    ]
                    return True
        return False

    def _update(self) -> None:
        """Moves steps from _pending to _executable / _pending_skip / _pending_retry
        as a function of what has been _completed.
        """
        new_steps_to_execute: list[str] = []
        new_steps_to_skip: list[str] = []
        new_steps_to_abandon: list[str] = []

        successful_or_skipped_steps = self._success | self._skipped
        failed_or_abandoned_steps = self._failed | self._abandoned
        resolved_steps = self._success | self._skipped | self._failed | self._abandoned

        if self._new_dynamic_mappings:
            new_step_deps = self._plan.resolve(self._completed_dynamic_outputs)
            for step_key, deps in new_step_deps.items():
                self._pending[step_key] = deps

            self._new_dynamic_mappings = False

        for step_key, depends_on_steps in self._pending.items():
            if depends_on_steps.issubset(resolved_steps):
                if self._should_skip_step(step_key, successful_or_skipped_steps):
                    new_steps_to_skip.append(step_key)
                elif depends_on_steps.intersection(failed_or_abandoned_steps):
                    new_steps_to_abandon.append(step_key)
                else:
                    new_steps_to_execute.append(step_key)

        for key in new_steps_to_execute:
            self._executable.append(key)
            del self._pending[key]

        for key in new_steps_to_skip:
            self._pending_skip.append(key)
            del self._pending[key]

        for key in new_steps_to_abandon:
            self._pending_abandon.append(key)
            del self._pending[key]

        ready_to_retry = []
        tick_time = time.time()
        for key, at_time in self._waiting_to_retry.items():
            if tick_time >= at_time:
                ready_to_retry.append(key)

        for key in ready_to_retry:
            self._executable.append(key)
            del self._waiting_to_retry[key]

    def sleep_interval(self):
        now = time.time()
        intervals = []
        if self._waiting_to_retry:
            for t in self._waiting_to_retry.values():
                intervals.append(t - now)
        if (
            self._instance_concurrency_context
            and self._instance_concurrency_context.has_pending_claims()
        ):
            intervals.append(
                self._instance_concurrency_context.interval_to_next_pending_claim_check()
            )
        if intervals:
            return min(intervals)

        return 0

    def sleep_til_ready(self) -> None:
        sleep_amt = self.sleep_interval()
        if sleep_amt > 0:
            time.sleep(sleep_amt)

    def get_next_step(self) -> Optional[ExecutionStep]:
        check.invariant(not self.is_complete, "Can not call get_next_step when is_complete is True")

        steps = self.get_steps_to_execute(limit=1)

        if not steps:
            return None

        return steps[0]

    def get_step_by_key(self, step_key: str) -> ExecutionStep:
        step = self._plan.get_step_by_key(step_key)
        return check.inst(step, ExecutionStep)

    def get_steps_to_execute(
        self,
        limit: Optional[int] = None,
    ) -> Sequence[ExecutionStep]:
        check.invariant(
            self._context_guard,
            "ActiveExecution must be used as a context manager",
        )
        check.opt_int_param(limit, "limit")

        self._update()

        steps = sorted(
            [self.get_step_by_key(key) for key in self._executable],
            key=self._sort_key_fn,
        )

        run_scoped_concurrency_limits_counter = None
        if self._tag_concurrency_limits:
            in_flight_steps = [self.get_step_by_key(key) for key in self._in_flight]
            run_scoped_concurrency_limits_counter = TagConcurrencyLimitsCounter(
                self._tag_concurrency_limits,
                in_flight_steps,
            )

        batch: list[ExecutionStep] = []

        for step in steps:
            if limit is not None and len(batch) >= limit:
                break

            if (
                self._max_concurrent is not None
                and len(batch) + len(self._in_flight) >= self._max_concurrent
            ):
                break

            if run_scoped_concurrency_limits_counter:
                if run_scoped_concurrency_limits_counter.is_blocked(step):
                    continue

            if run_scoped_concurrency_limits_counter:
                run_scoped_concurrency_limits_counter.update_counters_with_launched_item(step)

            # fallback to fetching from tags for backwards compatibility
            concurrency_key = _pool_key_for_step(step)
            if concurrency_key and self._instance_concurrency_context:
                try:
                    step_priority = int(step.tags.get(PRIORITY_TAG, 0))
                except ValueError:
                    step_priority = 0

                if not self._instance_concurrency_context.claim(
                    concurrency_key, step.key, step_priority
                ):
                    continue

            batch.append(step)

        for step in batch:
            self._in_flight.add(step.key)
            self._executable.remove(step.key)
            self._prep_for_dynamic_outputs(step)

        return batch

    def get_steps_to_skip(self) -> Sequence[ExecutionStep]:
        self._update()

        steps = []
        steps_to_skip = list(self._pending_skip)
        for key in steps_to_skip:
            step = self.get_step_by_key(key)
            steps.append(step)
            self._in_flight.add(key)
            self._pending_skip.remove(key)
            self._gathering_dynamic_outputs  # noqa: B018
            self._skip_for_dynamic_outputs(step)

        return sorted(steps, key=self._sort_key_fn)

    def get_steps_to_abandon(self) -> Sequence[ExecutionStep]:
        self._update()

        steps = []
        steps_to_abandon = list(self._pending_abandon)
        for key in steps_to_abandon:
            steps.append(self.get_step_by_key(key))
            self._in_flight.add(key)
            self._pending_abandon.remove(key)

        return sorted(steps, key=self._sort_key_fn)

    def plan_events_iterator(
        self, job_context: Union[PlanExecutionContext, PlanOrchestrationContext]
    ) -> Iterator[DagsterEvent]:
        """Process all steps that can be skipped and abandoned."""
        steps_to_skip = self.get_steps_to_skip()
        while steps_to_skip:
            for step in steps_to_skip:
                step_context = job_context.for_step(step)
                step_context.log.info(
                    f"Skipping step {step.key} due to skipped dependencies:"
                    f" {self._skipped_deps[step.key]}."
                )
                yield DagsterEvent.step_skipped_event(step_context)

                self.mark_skipped(step.key)

            steps_to_skip = self.get_steps_to_skip()

        steps_to_abandon = self.get_steps_to_abandon()
        while steps_to_abandon:
            for step in steps_to_abandon:
                step_context = job_context.for_step(step)
                failed_inputs: list[str] = []
                for step_input in step.step_inputs:
                    failed_inputs.extend(self._failed.intersection(step_input.dependency_keys))

                abandoned_inputs: list[str] = []
                for step_input in step.step_inputs:
                    abandoned_inputs.extend(
                        self._abandoned.intersection(step_input.dependency_keys)
                    )

                step_context.log.error(
                    "Dependencies for step {step}{fail_str}{abandon_str}. Not executing.".format(
                        step=step.key,
                        fail_str=f" failed: {failed_inputs}" if failed_inputs else "",
                        abandon_str=(
                            f" were not executed: {abandoned_inputs}" if abandoned_inputs else ""
                        ),
                    )
                )
                self.mark_abandoned(step.key)

            steps_to_abandon = self.get_steps_to_abandon()

    def mark_failed(self, step_key: str) -> None:
        self._failed.add(step_key)
        self._mark_complete(step_key)

    def mark_success(self, step_key: str) -> None:
        self._success.add(step_key)
        self._mark_complete(step_key)
        self._resolve_any_dynamic_outputs(step_key)

    def mark_skipped(self, step_key: str) -> None:
        self._skipped.add(step_key)
        self._mark_complete(step_key)
        self._resolve_any_dynamic_outputs(step_key)

    def mark_abandoned(self, step_key: str) -> None:
        self._abandoned.add(step_key)
        self._mark_complete(step_key)

    def mark_interrupted(self) -> None:
        self._interrupted = True

    def check_for_interrupts(self) -> bool:
        return pop_captured_interrupt()

    def mark_up_for_retry(self, step_key: str, at_time: Optional[float] = None) -> None:
        check.invariant(
            not self._retry_mode.disabled,
            f"Attempted to mark {step_key} as up for retry but retries are disabled",
        )
        check.opt_float_param(at_time, "at_time")

        # if retries are enabled - queue this back up
        if self._retry_mode.enabled:
            if at_time:
                self._waiting_to_retry[step_key] = at_time
            else:
                self._pending[step_key] = self._plan.get_executable_step_deps()[step_key]

        elif self._retry_mode.deferred:
            # do not attempt to execute again
            self._abandoned.add(step_key)

        self._retry_state.mark_attempt(step_key)

        self._mark_complete(step_key)

    def _mark_complete(self, step_key: str) -> None:
        check.invariant(
            step_key in self._in_flight,
            f"Attempted to mark step {step_key} as complete that was not known to be in flight",
        )
        self._in_flight.remove(step_key)

    def handle_event(self, dagster_event: DagsterEvent) -> None:
        check.inst_param(dagster_event, "dagster_event", DagsterEvent)

        step_key = cast(str, dagster_event.step_key)
        if dagster_event.is_step_failure:
            self.mark_failed(step_key)
            if self._instance_concurrency_context:
                self._instance_concurrency_context.free_step(step_key)
        elif dagster_event.is_resource_init_failure:
            # Resources are only initialized without a step key in the
            # in-process case, and resource initalization happens before the
            # ActiveExecution object is created.
            check.invariant(
                dagster_event.step_key is not None,
                "Resource init failure was reported during execution without a step key.",
            )
            self.mark_failed(step_key)
            if self._instance_concurrency_context:
                self._instance_concurrency_context.free_step(step_key)
        elif dagster_event.is_step_success:
            self.mark_success(step_key)
            if self._instance_concurrency_context:
                self._instance_concurrency_context.free_step(step_key)
        elif dagster_event.is_step_skipped:
            # Skip events are generated by this class. They should not be sent via handle_event
            raise DagsterInvariantViolationError(
                f"Step {step_key} was reported as skipped from outside the ActiveExecution."
            )
        elif dagster_event.is_step_up_for_retry:
            self.mark_up_for_retry(
                step_key,
                (
                    time.time() + dagster_event.step_retry_data.seconds_to_wait
                    if dagster_event.step_retry_data.seconds_to_wait
                    else None
                ),
            )
            if self._instance_concurrency_context:
                self._instance_concurrency_context.free_step(step_key)
        elif dagster_event.is_successful_output:
            event_specific_data = cast(StepOutputData, dagster_event.event_specific_data)
            self.mark_step_produced_output(event_specific_data.step_output_handle)
            if dagster_event.step_output_data.step_output_handle.mapping_key:
                check.not_none(
                    self._gathering_dynamic_outputs[step_key][
                        dagster_event.step_output_data.step_output_handle.output_name
                    ],
                ).append(dagster_event.step_output_data.step_output_handle.mapping_key)

    def verify_complete(self, job_context: IPlanContext, step_key: str) -> None:
        """Ensure that a step has reached a terminal state, if it has not mark it as an unexpected failure."""
        if step_key in self._in_flight:
            job_context.log.error(
                f"Step {step_key} finished without success or failure event. Downstream steps will not"
                " execute."
            )
            self.mark_unknown_state(step_key)

    # factored out for test
    def mark_unknown_state(self, step_key: str) -> None:
        # note the step so that we throw upon plan completion
        self._unknown_state.add(step_key)
        # mark as abandoned so downstream tasks do not execute
        self.mark_abandoned(step_key)

    # factored out for test
    def mark_step_produced_output(self, step_output_handle: StepOutputHandle) -> None:
        self._step_outputs.add(step_output_handle)

    @property
    def is_complete(self) -> bool:
        return (
            len(self._pending) == 0
            and len(self._in_flight) == 0
            and len(self._executable) == 0
            and len(self._pending_skip) == 0
            and len(self._pending_retry) == 0
            and len(self._pending_abandon) == 0
            and len(self._waiting_to_retry) == 0
            and (
                not self._instance_concurrency_context
                or not self._instance_concurrency_context.has_pending_claims()
            )
        )

    @property
    def retry_state(self) -> RetryState:
        return self._retry_state

    @property
    def has_in_flight_steps(self) -> bool:
        return len(self._in_flight) > 0

    def get_known_state(self) -> KnownExecutionState:
        return KnownExecutionState(
            previous_retry_attempts=self._retry_state.snapshot_attempts(),
            dynamic_mappings=dict(self._completed_dynamic_outputs),
            ready_outputs=self._step_outputs,
            step_output_versions=self._plan.known_state.step_output_versions,
            parent_state=self._plan.known_state.parent_state,
        )

    def _prep_for_dynamic_outputs(self, step: ExecutionStep):
        dyn_outputs = [step_out for step_out in step.step_outputs if step_out.is_dynamic]
        if dyn_outputs:
            self._gathering_dynamic_outputs[step.key] = {out.name: [] for out in dyn_outputs}

    def _skip_for_dynamic_outputs(self, step: ExecutionStep):
        dyn_outputs = [step_out for step_out in step.step_outputs if step_out.is_dynamic]
        if dyn_outputs:
            # place None to indicate the dynamic output was skipped, different than having 0 entries
            self._gathering_dynamic_outputs[step.key] = {out.name: None for out in dyn_outputs}

    def _resolve_any_dynamic_outputs(self, step_key: str) -> None:
        if step_key in self._gathering_dynamic_outputs:
            step = self.get_step_by_key(step_key)
            completed_mappings: dict[str, Optional[Sequence[str]]] = {}
            for output_name, mappings in self._gathering_dynamic_outputs[step_key].items():
                # if no dynamic outputs were returned and the output was marked is_required=False
                # set to None to indicate a skip should occur
                if not mappings and not step.step_output_dict[output_name].is_required:
                    completed_mappings[output_name] = None
                else:
                    completed_mappings[output_name] = mappings

            self._completed_dynamic_outputs[step_key] = completed_mappings
            self._new_dynamic_mappings = True

    def rebuild_from_events(
        self, dagster_events: Sequence[DagsterEvent]
    ) -> Sequence[ExecutionStep]:
        """Replay events to rebuild the execution state and continue after a failure.

        Returns a list of steps that are possibly in flight. Current status of the event log implies
        that the previous run worker might have crashed before launching these steps, or it may have
        launched them but they have yet to report a STEP_START event.
        """
        self.get_steps_to_execute()

        for event in dagster_events:
            self.handle_event(event)
            self.get_steps_to_execute()

        return [self.get_step_by_key(step_key) for step_key in self._in_flight]

    def concurrency_event_iterator(
        self, plan_context: Union[PlanExecutionContext, PlanOrchestrationContext]
    ) -> Iterator[DagsterEvent]:
        if not self._instance_concurrency_context:
            return

        pending_claims = self._instance_concurrency_context.pending_claim_steps()
        for step_key in pending_claims:
            last_messaged_timestamp = self._messaged_concurrency_slots.get(step_key)
            if (
                not last_messaged_timestamp
                or time.time() - last_messaged_timestamp > CONCURRENCY_CLAIM_MESSAGE_INTERVAL
            ):
                step = self.get_step_by_key(step_key)
                step_context = plan_context.for_step(step)
                pool = check.inst(_pool_key_for_step(step), str)
                self._messaged_concurrency_slots[step_key] = time.time()
                is_initial_message = last_messaged_timestamp is None
                yield DagsterEvent.step_concurrency_blocked(
                    step_context, pool, initial=is_initial_message
                )



from collections.abc import Mapping
from typing import Any, Dict  # noqa: F401, UP035

from dagster._core.definitions.tags.tag_set import NamespacedTagSet as NamespacedTagSet
from dagster._core.storage.tags import KIND_PREFIX


def has_kind(tags: Mapping[str, Any], kind: str) -> bool:
    return build_kind_tag_key(kind) in tags


def build_kind_tag_key(kind: str) -> str:
    return f"{KIND_PREFIX}{kind}"


def build_kind_tag(kind: str) -> dict[str, Any]:
    return {build_kind_tag_key(kind): ""}



# flake8: noqa
# type: ignore
# Generated from AssetSelection.g4 by ANTLR 4.13.2
from antlr4 import *
from .AssetSelectionParser import AssetSelectionParser
# This class defines a complete generic visitor for a parse tree produced by AssetSelectionParser.


class AssetSelectionVisitor(ParseTreeVisitor):
    # Visit a parse tree produced by AssetSelectionParser#start.
    def visitStart(self, ctx: AssetSelectionParser.StartContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#UpTraversalExpression.
    def visitUpTraversalExpression(self, ctx: AssetSelectionParser.UpTraversalExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#AndExpression.
    def visitAndExpression(self, ctx: AssetSelectionParser.AndExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#AllExpression.
    def visitAllExpression(self, ctx: AssetSelectionParser.AllExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#TraversalAllowedExpression.
    def visitTraversalAllowedExpression(
        self, ctx: AssetSelectionParser.TraversalAllowedExpressionContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#DownTraversalExpression.
    def visitDownTraversalExpression(
        self, ctx: AssetSelectionParser.DownTraversalExpressionContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#NotExpression.
    def visitNotExpression(self, ctx: AssetSelectionParser.NotExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#OrExpression.
    def visitOrExpression(self, ctx: AssetSelectionParser.OrExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#UpAndDownTraversalExpression.
    def visitUpAndDownTraversalExpression(
        self, ctx: AssetSelectionParser.UpAndDownTraversalExpressionContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#AttributeExpression.
    def visitAttributeExpression(self, ctx: AssetSelectionParser.AttributeExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#FunctionCallExpression.
    def visitFunctionCallExpression(self, ctx: AssetSelectionParser.FunctionCallExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#ParenthesizedExpression.
    def visitParenthesizedExpression(
        self, ctx: AssetSelectionParser.ParenthesizedExpressionContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#upTraversal.
    def visitUpTraversal(self, ctx: AssetSelectionParser.UpTraversalContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#downTraversal.
    def visitDownTraversal(self, ctx: AssetSelectionParser.DownTraversalContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#functionName.
    def visitFunctionName(self, ctx: AssetSelectionParser.FunctionNameContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#KeyExpr.
    def visitKeyExpr(self, ctx: AssetSelectionParser.KeyExprContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#TagAttributeExpr.
    def visitTagAttributeExpr(self, ctx: AssetSelectionParser.TagAttributeExprContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#OwnerAttributeExpr.
    def visitOwnerAttributeExpr(self, ctx: AssetSelectionParser.OwnerAttributeExprContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#GroupAttributeExpr.
    def visitGroupAttributeExpr(self, ctx: AssetSelectionParser.GroupAttributeExprContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#KindAttributeExpr.
    def visitKindAttributeExpr(self, ctx: AssetSelectionParser.KindAttributeExprContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#ColumnAttributeExpr.
    def visitColumnAttributeExpr(self, ctx: AssetSelectionParser.ColumnAttributeExprContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#TableNameAttributeExpr.
    def visitTableNameAttributeExpr(self, ctx: AssetSelectionParser.TableNameAttributeExprContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#ColumnTagAttributeExpr.
    def visitColumnTagAttributeExpr(self, ctx: AssetSelectionParser.ColumnTagAttributeExprContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#CodeLocationAttributeExpr.
    def visitCodeLocationAttributeExpr(
        self, ctx: AssetSelectionParser.CodeLocationAttributeExprContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#ChangedInBranchAttributeExpr.
    def visitChangedInBranchAttributeExpr(
        self, ctx: AssetSelectionParser.ChangedInBranchAttributeExprContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#value.
    def visitValue(self, ctx: AssetSelectionParser.ValueContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by AssetSelectionParser#keyValue.
    def visitKeyValue(self, ctx: AssetSelectionParser.KeyValueContext):
        return self.visitChildren(ctx)


del AssetSelectionParser



"""add backfill_id column to runs table

Revision ID: 63d7a8ec641a
Revises: 284a732df317
Create Date: 2024-10-18 15:56:30.002452

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "63d7a8ec641a"
down_revision = "284a732df317"
branch_labels = None
depends_on = None


def upgrade():
    inspector = sa.inspect(op.get_bind())
    has_tables = inspector.get_table_names()

    if "runs" in has_tables:
        columns = [x.get("name") for x in inspector.get_columns("runs")]
        if "backfill_id" not in columns:
            op.add_column("runs", sa.Column("backfill_id", sa.String(255), nullable=True))


def downgrade():
    inspector = sa.inspect(op.get_bind())
    has_tables = inspector.get_table_names()
    if "runs" in has_tables:
        columns = [x.get("name") for x in inspector.get_columns("runs")]
        if "backfill_id" in columns:
            op.drop_column("runs", "backfill_id")
